import unittest
from typing import Any
from unittest.mock import MagicMock, patch

from loguru import logger
from requests.exceptions import HTTPError, RequestException

from girsh.core import repos


class SubstringMatcher:
    def __init__(self, containing: str) -> None:
        self.containing: str = containing.lower()

    def __eq__(self, other: Any) -> Any:
        return other.lower().find(self.containing) > -1

    def __unicode__(self) -> str:
        return f'a string containing "{self.containing}"'

    def __str__(self) -> str:
        return self.__unicode__().encode("utf-8").decode("utf-8")

    __repr__ = __unicode__


# -------------------------------
# Dummy classes used for testing
# -------------------------------
class DummyResponse:
    def __init__(self, json_data: Any, raise_for_status: bool = False) -> None:
        self._json = json_data
        self._raise_exc = raise_for_status
        self.headers: dict[str, str] = {}

    def raise_for_status(self) -> None:
        if self._raise_exc:
            raise HTTPError("status")

    def json(self) -> Any:
        return self._json


# -------------------------------
# Tests for fetch_release_info
# -------------------------------
class TestFetchReleaseInfo(unittest.TestCase):
    def setUp(self) -> None:
        # Disable loguru logger output

        logger.remove()
        self.repo_version = "v1.0"

    def test_fetch_release_info_valid(self) -> None:
        dummy_json = {"tag_name": self.repo_version, "url": "dummy"}
        with patch("requests.get", return_value=DummyResponse(dummy_json)):
            info: dict[Any, Any] | None = repos.fetch_release_info(
                "owner/repo",
                self.repo_version,
                False,
            )
        self.assertEqual(info, dummy_json)

    def test_fetch_release_info_invalid_data(self) -> None:
        dummy_json = ["not", "a", "dict"]
        with patch("requests.get", return_value=DummyResponse(dummy_json)):
            info: dict[Any, Any] | None = repos.fetch_release_info(
                "owner/repo",
                self.repo_version,
                False,
            )
        self.assertIsNone(info)

    def test_fetch_release_info_status_exception(self) -> None:
        dummy_json = {"tag_name": self.repo_version, "url": "dummy"}
        with (
            patch("requests.get", return_value=DummyResponse(dummy_json, raise_for_status=True)),
            patch.object(logger, "error") as mock_logger_error,
        ):
            info: dict[Any, Any] | None = repos.fetch_release_info(
                "owner/repo",
                self.repo_version,
                False,
            )
            mock_logger_error.assert_called_once_with(SubstringMatcher(containing="status"))
        self.assertIsNone(info)

    def test_fetch_release_info_general_exception(self) -> None:
        with (
            patch("requests.get", side_effect=RequestException("error")),
            patch.object(logger, "error") as mock_logger_error,
        ):
            info: dict[Any, Any] | None = repos.fetch_release_info(
                "owner/repo",
                self.repo_version,
                False,
            )
            mock_logger_error.assert_called_once_with(SubstringMatcher(containing="error"))
        self.assertIsNone(info)


# -------------------------------
# Tests for is_new_version
# -------------------------------
class TestIsNewVersion(unittest.TestCase):
    def test_no_installed_info(self) -> None:
        self.assertTrue(repos.is_new_version(None, "v1", False))

    def test_different_version(self) -> None:
        self.assertTrue(repos.is_new_version("v0.9", "v1", False))

    def test_same_version_no_reinstall(self) -> None:
        self.assertEqual(repos.is_new_version("v1", "v1", False), repos.RepoResult.skipped)

    def test_same_version_with_reinstall(self) -> None:
        self.assertEqual(repos.is_new_version("v1", "v1", True), repos.RepoResult.installed)


# -------------------------------
# Tests for check_repo_release
# -------------------------------
class TestCheckRepoRelease(unittest.TestCase):
    # def setUp(self) -> None:
    #     # logger.remove()
    #     self.repo = "dummy/repo"
    #     self.version = "v1.0"

    @patch("girsh.core.repos.logger")
    @patch("girsh.core.repos.fetch_release_info", return_value=None)
    def test_check_repo_release_no_release_info(
        self, mock_fetch_release_info: MagicMock, mock_logger: MagicMock
    ) -> None:
        self.assertEqual(
            repos.check_repo_release(repo="dummy", target_version=None, current_version=None, reinstall=False),
            (repos.RepoResult.install_failed, {}),
        )
        mock_logger.error.assert_called_once_with("No valid release info received")

    @patch("girsh.core.repos.logger")
    @patch("girsh.core.repos.fetch_release_info", return_value={"tag_name": "v1.0"})
    @patch("girsh.core.repos.is_new_version", return_value=repos.RepoResult.installed)
    def test_check_repo_release_new_version(
        self, mock_is_new_version: MagicMock, mock_fetch_release_info: MagicMock, mock_logger: MagicMock
    ) -> None:
        self.assertEqual(
            repos.check_repo_release(repo="dummy", target_version=None, current_version=None, reinstall=False),
            (repos.RepoResult.installed, {"tag_name": "v1.0"}),
        )
        mock_logger.error.assert_not_called()

    @patch("girsh.core.repos.logger")
    @patch("girsh.core.repos.fetch_release_info", return_value={"tag_name": "v1.0"})
    @patch("girsh.core.repos.is_new_version", return_value=repos.RepoResult.skipped)
    def test_check_repo_release_skipped(
        self, mock_is_new_version: MagicMock, mock_fetch_release_info: MagicMock, mock_logger: MagicMock
    ) -> None:
        self.assertEqual(
            repos.check_repo_release(repo="dummy", target_version=None, current_version=None, reinstall=False),
            (repos.RepoResult.skipped, {}),
        )
        mock_logger.info.assert_called_once_with("dummy: No newer version than v1.0 found, skipping.")
        mock_logger.error.assert_not_called()


if __name__ == "__main__":
    unittest.main()



import contextlib
import sys
import tempfile
import unittest
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any
from unittest.mock import patch

# Import the module under test as specified.
from girsh import girsh


# --- Dummy CLI arguments ---
@dataclass
class DummyArgs:
    edit: bool = False
    clean: bool = False
    verbose: bool = False
    show: bool = False
    uninstall: bool = False
    uninstall_all: bool = False
    dry_run: bool = False
    reinstall: bool = False
    config: str = "dummy_config.yaml"
    system: bool = False


# --- Dummy general config class ---
class DummyGeneral:
    pass


class DummyRepoResult(Enum):
    # Installation status changed
    result1 = 1
    result2 = 2
    fail1 = 3
    fail2 = 4


DUMMY_EDIT_RETURN = 42
DUMMY_SHOW_RETURN = 43
DUMMY_ELEVATED_EXIT = 44
DUMMY_REPO_CONFIG = {"repo1": "config1", "repo2": "config2"}


def dummy_elevate_privileges() -> None:
    sys.exit(DUMMY_ELEVATED_EXIT)


def dummy_edit_config(config: str) -> int:
    return DUMMY_EDIT_RETURN


def dummy_clean_downloads_folder(download_dir: Path) -> int:
    return 0


def dummy_load_yaml_config(config: str) -> tuple[DummyGeneral, dict]:
    # Do nothing for testing.
    general: Any = DummyGeneral()
    general.installed_file = "dummy_installed_file"
    general.download_dir = "dummy_download_dir"
    return general, DUMMY_REPO_CONFIG


def dummy_load_yaml_config_empty(config: str) -> tuple[DummyGeneral, dict]:
    # Do nothing for testing.
    general: Any = DummyGeneral()
    general.installed_file = "dummy_installed_file"
    return general, {}


def dummy_load_installed_empty(installed_file: str) -> dict[Any, Any]:
    return {}


def dummy_show_installed(installed: dict[Any, Any], repositories: dict) -> int:
    return DUMMY_SHOW_RETURN


def dummy_uninstall(repositories: dict, installed: dict[Any, Any], dry_run: bool) -> dict:
    return defaultdict(int)


def dummy_uninstall_all(
    installed_file: Path, installed: dict[Any, Any], repositories: dict, dry_run: bool = False
) -> int:
    installed.clear()
    return 0


def dummy_process_repositories(
    repositories: dict, general: Any, installed: dict[Any, Any], reinstall: bool, dry_run: bool
) -> tuple[dict[str, dict], dict]:
    for repo in repositories:
        installed[repo] = {"processed": True}
    return installed, defaultdict(int)


def dummy_process_repositories_no_installed(
    repositories: dict, general: Any, installed: dict[Any, Any], reinstall: bool, dry_run: bool
) -> tuple[dict[str, dict], dict]:
    return {}, {}


def dummy_save_installed(installed_file: str, installed: dict[Any, Any]) -> int:
    return 0


# --- Test class for girsh.main() ---


class TestMain(unittest.TestCase):
    def setUp(self) -> None:
        # Disable loguru logger output.
        from loguru import logger

        logger.remove()

        self.dummy_general_config: Any = DummyGeneral()
        self.dummy_general_config.installed_file = "dummy_installed_file"
        # Use a temporary directory for downloads.
        self.dummy_general_config.download_dir = Path(tempfile.gettempdir()) / "dummy_download_dir"
        self.dummy_general_config.download_dir.mkdir(exist_ok=True)
        # Create a dummy repository configuration.
        self.dummy_repositories_config = {"dummy_repo": "dummy_repo_config"}
        # Create a temporary directory to simulate file operations if needed.

        self.temp_dir: Path = Path(tempfile.mkdtemp())

    def tearDown(self) -> None:
        # Clean up temporary directory.
        if self.temp_dir.exists():
            with contextlib.suppress(Exception):
                self.temp_dir.rmdir()

    def test_elevate_privileges(self) -> None:
        dummy_args = DummyArgs(system=True)
        with (
            patch("girsh.girsh.get_arguments", return_value=dummy_args),
            patch("girsh.girsh.elevate_privileges", side_effect=dummy_elevate_privileges),
            patch("sys.exit", side_effect=SystemExit) as mock_exit,
        ):
            with self.assertRaises(SystemExit):
                girsh.main()
            mock_exit.assert_called_once_with(DUMMY_ELEVATED_EXIT)

    def test_edit_branch(self) -> None:
        dummy_args = DummyArgs(edit=True, config="config.yaml")
        with (
            patch("girsh.girsh.get_arguments", return_value=dummy_args),
            patch("girsh.girsh.edit_config", side_effect=dummy_edit_config) as mock_edit,
        ):
            ret: int = girsh.main()
            self.assertEqual(ret, DUMMY_EDIT_RETURN)
            mock_edit.assert_called_once_with("config.yaml")

    def test_clean_branch(self) -> None:
        dummy_args = DummyArgs(clean=True, config="config.yaml")
        with (
            patch("girsh.girsh.get_arguments", return_value=dummy_args),
            patch("girsh.girsh.load_yaml_config", side_effect=dummy_load_yaml_config),
            patch("girsh.girsh.clean_downloads_folder", side_effect=dummy_clean_downloads_folder) as mock_clean,
        ):
            ret: int = girsh.main()
            self.assertEqual(ret, 0)
            mock_clean.assert_called_once()

    def test_verbose_default_branch(self) -> None:
        dummy_args = DummyArgs(verbose=True, config="config.yaml", reinstall=False, dry_run=False)
        dummy_installed_file: str = "dummy_installed_file"

        repositories = {"repo1": "config1", "repo2": "config2"}
        dummy_installed: dict[Any, Any] = repositories
        with (
            patch("girsh.girsh.get_arguments", return_value=dummy_args),
            patch("girsh.girsh.load_yaml_config", side_effect=dummy_load_yaml_config) as _mock_load_yaml,
            patch("girsh.girsh.load_installed", side_effect=dummy_load_installed_empty) as _mock_load_installed,
            patch("girsh.girsh.process_repositories", side_effect=dummy_process_repositories) as _mock_process,
            patch("girsh.girsh.save_installed", side_effect=dummy_save_installed) as mock_save_installed,
            patch("girsh.girsh.logger.add", lambda *args, **kwargs: None),
            patch("girsh.girsh.logger.debug", lambda *args, **kwargs: None),
        ):
            ret: int = girsh.main()
        self.assertEqual(ret, 0)
        for repo in repositories:
            self.assertIn(repo, dummy_installed)
        mock_save_installed.assert_called_once_with(
            dummy_installed_file, {"repo1": {"processed": True}, "repo2": {"processed": True}}
        )

    def test_process_repositories_none_installed(self) -> None:
        dummy_args = DummyArgs(show=False, config="config.yaml")
        dummy_installed_data: dict[Any, Any] = {"a": {"path": "/path", "binary": "bin"}}
        with (
            patch("girsh.girsh.get_arguments", return_value=dummy_args),
            patch("girsh.girsh.load_yaml_config", side_effect=dummy_load_yaml_config),
            patch("girsh.girsh.load_installed", return_value=(dummy_installed_data)),
            patch("girsh.girsh.uninstall", side_effect=dummy_uninstall),
            patch("girsh.girsh.show_installed", side_effect=dummy_show_installed),
            patch(
                "girsh.girsh.process_repositories", side_effect=dummy_process_repositories_no_installed
            ) as mock_process_repositories,
        ):
            ret: int = girsh.main()
            self.assertEqual(ret, 3)
            mock_process_repositories.assert_called()

    def test_show_installed(self) -> None:
        dummy_args = DummyArgs(show=True, config="config.yaml")
        dummy_installed_data: dict[Any, Any] = {"a": 1}
        # dummy_repositories: dict = {}
        with (
            patch("girsh.girsh.get_arguments", return_value=dummy_args),
            patch("girsh.girsh.load_yaml_config", side_effect=dummy_load_yaml_config),
            patch("girsh.girsh.load_installed", return_value=(dummy_installed_data)),
            patch("girsh.girsh.show_installed", side_effect=dummy_show_installed) as mock_show,
        ):
            ret: int = girsh.main()
            self.assertEqual(ret, DUMMY_SHOW_RETURN)
            mock_show.assert_called_once_with(dummy_installed_data, DUMMY_REPO_CONFIG)

    def test_uninstall_branch(self) -> None:
        dummy_args = DummyArgs(uninstall=True, dry_run=True, config="config.yaml")
        dummy_installed: dict[Any, Any] = {"a": 1}
        dummy_repositories: dict = {}
        with (
            patch("girsh.girsh.get_arguments", return_value=dummy_args),
            patch("girsh.girsh.load_yaml_config", side_effect=dummy_load_yaml_config_empty),
            patch("girsh.girsh.load_installed", return_value=(dummy_installed)),
            patch("girsh.girsh.uninstall", side_effect=dummy_uninstall) as mock_uninstall,
            patch("girsh.girsh.save_installed", side_effect=dummy_save_installed) as mock_save,
        ):
            ret: int = girsh.main()
            self.assertEqual(ret, 0)
            mock_uninstall.assert_called_once_with(
                repositories=list(dummy_repositories), installed=dummy_installed, dry_run=True
            )
            mock_save.assert_called_once_with("dummy_installed_file", dummy_installed)

    def test_uninstall_all_installed(self) -> None:
        dummy_args = DummyArgs(uninstall_all=True, dry_run=True, config="config.yaml")
        dummy_installed: dict[Any, Any] = {}
        with (
            patch("girsh.girsh.get_arguments", return_value=dummy_args),
            patch("girsh.girsh.load_yaml_config", side_effect=dummy_load_yaml_config),
            patch("girsh.girsh.load_installed", return_value=(dummy_installed)),
            patch("girsh.girsh.uninstall") as mock_uninstall,
            patch("girsh.girsh.save_installed", side_effect=dummy_save_installed) as mock_save,
            patch("girsh.girsh.show_installed") as mock_show,
        ):
            ret: int = girsh.main()
            # If installed is empty, uninstall_all branch should not be called.
            self.assertEqual(ret, 0)
            mock_uninstall.assert_called_with(repositories=[], installed={}, dry_run=True)
            mock_show.assert_not_called()
            mock_save.assert_called_once_with("dummy_installed_file", dummy_installed)

    def test_uninstall_all_nonempty_installed(self) -> None:
        dummy_args = DummyArgs(uninstall_all=True, dry_run=False, config="config.yaml")
        dummy_installed: dict[Any, Any] = {"repo": {"tag": "v1"}}
        with (
            patch("girsh.girsh.get_arguments", return_value=dummy_args),
            patch("girsh.girsh.load_yaml_config", side_effect=dummy_load_yaml_config_empty),
            patch("girsh.girsh.load_installed", return_value=(dummy_installed)),
            patch("girsh.girsh.uninstall") as mock_uninstall,
            patch("girsh.girsh.save_installed", side_effect=dummy_save_installed) as mock_save,
        ):
            ret: int = girsh.main()
            self.assertEqual(ret, 0)
            mock_uninstall.assert_called_once_with(repositories=[], installed={"repo": {"tag": "v1"}}, dry_run=False)
            mock_save.assert_called_once_with("dummy_installed_file", dummy_installed)

    def test_default_branch_no_repositories(self) -> None:
        dummy_args = DummyArgs(config="config.yaml")

        dummy_installed: dict[Any, Any] = {}
        with (
            patch("girsh.girsh.get_arguments", return_value=dummy_args),
            patch("girsh.girsh.load_yaml_config", side_effect=dummy_load_yaml_config_empty),
            patch("girsh.girsh.load_installed", return_value=(dummy_installed)),
            patch("girsh.girsh.save_installed", side_effect=dummy_save_installed) as mock_save,
        ):
            ret: int = girsh.main()
            self.assertEqual(ret, 0)
            mock_save.assert_called()


class ElevatePrivilegesTest(unittest.TestCase):
    @patch("os.geteuid", return_value=0)
    def test_elevate_privileges_already_root(self, mock_geteuid: unittest.mock.Mock) -> None:
        with patch("builtins.print") as mock_print, patch("os.execvp") as mock_execvp:
            girsh.elevate_privileges()
            mock_print.assert_not_called()
            mock_execvp.assert_not_called()

    @patch("os.geteuid", return_value=1000)
    def test_elevate_privileges_not_root(self, mock_geteuid: unittest.mock.Mock) -> None:
        with patch("builtins.print") as mock_print, patch("os.execvp") as mock_execvp:
            girsh.elevate_privileges()
            mock_print.assert_called_once_with("Re-running with elevated privileges...")
            command = ["sudo", sys.executable, *sys.argv]
            mock_execvp.assert_called_once_with("/usr/bin/sudo", command)


class ShowSummaryTest(unittest.TestCase):
    @patch("girsh.girsh.logger")
    def test_show_summary_with_data(self, mock_logger: unittest.mock.Mock) -> None:
        install_summary = {DummyRepoResult.result1: 3, DummyRepoResult.result2: 1}
        uninstall_summary = {DummyRepoResult.fail1: 2, DummyRepoResult.fail2: 1}

        girsh.show_summary(install_summary, uninstall_summary)  # type: ignore[arg-type]

        mock_logger.success.assert_any_call("===============================")
        mock_logger.success.assert_any_call("Summary:")
        mock_logger.success.assert_any_call("  result1: 3")
        mock_logger.success.assert_any_call("  result2: 1")
        mock_logger.success.assert_any_call("  fail1: 2")
        mock_logger.success.assert_any_call("  fail2: 1")

    @patch("girsh.girsh.logger")
    def test_show_summary_empty(self, mock_logger: unittest.mock.Mock) -> None:
        install_summary: dict = {}
        uninstall_summary: dict = {}

        girsh.show_summary(install_summary, uninstall_summary)

        mock_logger.success.assert_any_call("===============================")
        mock_logger.success.assert_any_call("Summary:")


if __name__ == "__main__":
    unittest.main()



import tempfile
import unittest
from dataclasses import dataclass
from pathlib import Path
from typing import Any
from unittest.mock import MagicMock, patch

import yaml
from loguru import logger

from girsh.core import installed


# -------------------------------
# Dummy classes used for testing
# -------------------------------
@dataclass
class DummyRepo:
    comment: str


class TestInstalledModule(unittest.TestCase):
    def setUp(self) -> None:
        logger.remove()  # Remove any existing logger handlers to avoid duplicate logs.
        # Create a temporary directory and file to use as installed file.
        self.temp_dir: tempfile.TemporaryDirectory = tempfile.TemporaryDirectory()
        self.temp_path: Path = Path(self.temp_dir.name)
        self.installed_file: Path = self.temp_path / "installed.yaml"
        self.output_dir = self.temp_path / "output"
        self.download_dir = self.temp_path / "downloads"
        self.bin_base_folder = self.temp_path / "bin"
        self.package_base_folder = self.temp_path / "packages"

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    # --- Tests for load_installed ---

    def test_load_installed_normal(self) -> None:
        # Write valid YAML dict data to the installed file.
        expected_data: dict[str, dict[str, str]] = {"repo1": {"tag": "v1.0", "binary": "bin1"}}
        with self.installed_file.open("w") as f:
            yaml.safe_dump(expected_data, f)
        ret_data = installed.load_installed(installed_file=self.installed_file)
        self.assertEqual(ret_data, expected_data)

    def test_load_installed_non_dict(self) -> None:
        # Write valid YAML that is not a dict (e.g. a list).
        with self.installed_file.open("w") as f:
            yaml.safe_dump([1, 2, 3], f)
        ret_data = installed.load_installed(installed_file=self.installed_file)
        self.assertEqual(ret_data, {})

    def test_load_installed_file_not_exists(self) -> None:
        # Use a path that does not exist.
        non_existent: Path = self.temp_path / "nonexistent.yaml"
        if non_existent.exists():
            non_existent.unlink()
        ret_data = installed.load_installed(installed_file=non_existent)
        self.assertEqual(ret_data, {})

    @patch("girsh.core.installed.Path.mkdir")
    @patch("sys.exit", side_effect=SystemExit)
    def test_load_installed_permission_error_on_mkdir(self, mock_exit: MagicMock, mock_mkdir: MagicMock) -> None:
        # Simulate a PermissionError when creating the installed folder.
        mock_mkdir.side_effect = PermissionError("No permission")
        with patch.object(logger, "error") as mock_logger_error:
            with self.assertRaises(SystemExit):
                installed.load_installed(installed_file=self.installed_file)
            mock_logger_error.assert_called_once_with(
                f"Permission denied when creating installed folder '{self.installed_file.parent}'."
            )
        mock_exit.assert_called_once_with(1)

    def test_load_installed_permission_error_on_open(self) -> None:
        # Simulate PermissionError when opening the file for reading.
        self.installed_file.touch()  # Ensure the file exists.
        with (
            patch("girsh.core.installed.Path.exists", return_value=True),
            patch("pathlib.Path.open", side_effect=PermissionError("No permission")),
            patch.object(logger, "error") as mock_logger,
        ):
            ret_data = installed.load_installed(installed_file=self.installed_file)
            self.assertEqual(ret_data, {})
            mock_logger.assert_called_with(
                f"Warning: Permission denied when reading installed file '{self.installed_file}'. Proceeding with empty installed."
            )

    def test_load_installed_file_not_found_during_open(self) -> None:
        # Simulate FileNotFoundError when opening the file, even though exists() returns True.
        self.installed_file.touch()  # Ensure the file exists.
        with (
            patch("girsh.core.installed.Path.exists", return_value=False),
            # patch("girsh.core.installed.Path.open", side_effect=FileNotFoundError("Not found")),
            patch.object(logger, "info") as mock_logger_info,
        ):
            ret_data = installed.load_installed(installed_file=self.installed_file)
            self.assertEqual(ret_data, {})
            mock_logger_info.assert_called_with(
                f"Installed packages file '{self.installed_file}' not found. Proceeding with empty installed."
            )

    # --- Tests for save_installed ---

    def test_save_installed_success(self) -> None:
        data: dict[str, dict[str, str]] = {"repo2": {"tag": "v2.0", "binary": "bin2"}}
        ret: int = installed.save_installed(self.installed_file, data)
        self.assertEqual(ret, 0)
        # Verify that the file was written.
        with self.installed_file.open("r") as f:
            loaded: Any = yaml.safe_load(f)
        self.assertEqual(loaded, data)

    def test_save_installed_permission_error(self) -> None:
        with (
            patch("girsh.core.installed.Path.open", side_effect=PermissionError("No permission")),
            patch.object(logger, "error") as mock_logger_error,
        ):
            ret: int = installed.save_installed(self.installed_file, {"key": "value"})
            self.assertEqual(ret, 1)
            mock_logger_error.assert_called_once_with(
                f"Permission denied when writing to installed file '{self.installed_file}'."
            )

    def test_save_installed_file_not_found(self) -> None:
        with (
            patch("girsh.core.installed.Path.open", side_effect=FileNotFoundError("Not found")),
            patch.object(logger, "error") as mock_logger_error,
        ):
            ret: int = installed.save_installed(self.installed_file, {"key": "value"})
            self.assertEqual(ret, 1)
            mock_logger_error.assert_called_once_with(
                f"installed file '{self.installed_file}' not found and cannot be created."
            )

    # --- Tests for get_comment ---

    def test_get_comment_found(self) -> None:
        # Set up CONFIG to include a repository.
        repositories = {"test/repo": DummyRepo("A test repo")}
        comment: str = installed.get_comment("test/repo", repositories=repositories)  # type: ignore[arg-type]
        self.assertEqual(comment, "A test repo")

    def test_get_comment_not_found(self) -> None:
        with patch.object(logger, "debug") as mock_logger_debug:
            comment: str = installed.get_comment("nonexistent/repo", repositories={})
            self.assertEqual(comment, "Not found in config.")
            mock_logger_debug.assert_called_once_with(
                "The repository nonexistent/repo is not present in current config."
            )

    # --- Tests for show_installed ---

    def test_show_installed_empty(self) -> None:
        with patch.object(logger, "success") as mock_logger:
            ret: int = installed.show_installed({}, {})
            self.assertEqual(ret, 0)
            mock_logger.assert_called_with("Nothing installed via `girsh`")

    def test_show_installed_non_empty(self) -> None:
        # Set up CONFIG so that get_comment returns a meaningful comment.
        repositories = {"repo1": DummyRepo("Test comment")}
        data: dict[str, dict[str, str]] = {"repo1": {"binary": "binA", "tag": "v1.1"}}
        with patch.object(logger, "success") as mock_logger:
            ret: int = installed.show_installed(data, repositories)  # type: ignore[arg-type]
            self.assertEqual(ret, 0)
            # Verify that table components were logged.
            calls: list[Any] = [call.args[0] for call in mock_logger.call_args_list]
            self.assertTrue(any("Currently installed binaries:" in msg for msg in calls))
            # Expect at least one separator line that starts and ends with '+'.
            self.assertTrue(any(msg.startswith("+") and msg.endswith("+") for msg in calls))


if __name__ == "__main__":
    unittest.main()



import unittest
from typing import Any
from unittest.mock import MagicMock, patch

from loguru import logger

from girsh.core import utils


class SubstringMatcher:
    def __init__(self, containing: str) -> None:
        self.containing: str = containing.lower()

    def __eq__(self, other: Any) -> Any:
        return other.lower().find(self.containing) > -1

    def __unicode__(self) -> str:
        return f'a string containing "{self.containing}"'

    def __str__(self) -> str:
        return self.__unicode__().encode("utf-8").decode("utf-8")

    __repr__ = __unicode__


# -------------------------------
# Tests for confirm
# -------------------------------
class TestConfirmYesNo(unittest.TestCase):
    def test_confirm_default_no_y(self) -> None:
        responses = ["y", "Y"]
        with patch("builtins.input", side_effect=responses):
            for _ in responses:
                self.assertTrue(utils.confirm_default_no("dummy"))

    def test_confirm_default_no_x(self) -> None:
        responses = ["n", "N", "x", ""]
        with patch("builtins.input", side_effect=responses):
            for _ in responses:
                self.assertFalse(utils.confirm_default_no("dummy"))

    def test_confirm_default_yes_n(self) -> None:
        responses = ["n", "N"]
        with patch("builtins.input", side_effect=responses):
            for _ in responses:
                self.assertFalse(utils.confirm_default_yes("dummy"))

    def test_confirm_default_yes_x(self) -> None:
        responses = ["y", "Y", "x", ""]
        with patch("builtins.input", side_effect=responses):
            for _ in responses:
                self.assertTrue(utils.confirm_default_yes("dummy"))


# -------------------------------
# Dummy classes used for testing
# -------------------------------
class DummyProcess:
    def __init__(self, name: str) -> None:
        self.info: dict[str, Any] = {"name": name, "cmdline": ["dummy", "command"], "pid": 42}

    def terminate(self) -> None:
        return

    def kill(self) -> None:
        return


# -------------------------------
# Tests for get_processes
# -------------------------------
class TestGetProcesses(unittest.TestCase):
    def test_get_processes(self) -> None:
        dummy1 = DummyProcess("my_prog")
        dummy2 = DummyProcess("other")
        with patch("psutil.process_iter", return_value=[dummy1, dummy2]):
            processes: list[DummyProcess] = utils.get_processes("my_prog")  # type: ignore[assignment]
        self.assertEqual(processes, [dummy1])


# -------------------------------
# Tests for stop_processes
# -------------------------------
class TestStopProcesses(unittest.TestCase):
    def setUp(self) -> None:
        # Disable loguru logger output

        logger.remove()

    def test_stop_processes_no_processes(self) -> None:
        with patch("girsh.core.utils.get_processes", return_value=[]):
            result: bool = utils.stop_processes("prog")
        self.assertTrue(result)

    def test_stop_processes_cancel_termination(self) -> None:
        dummy = DummyProcess("prog")
        with patch("girsh.core.utils.get_processes", return_value=[dummy]), patch("builtins.input", return_value="n"):
            result: bool = utils.stop_processes("prog")
        self.assertFalse(result)

    def test_stop_processes_success(self) -> None:
        dummy = DummyProcess("prog")
        with (
            patch("girsh.core.utils.get_processes", side_effect=[[dummy], []]),
            patch("builtins.input", return_value="y"),
        ):
            result: bool = utils.stop_processes("prog")
        self.assertTrue(result)

    def test_stop_processes_fail_then_cancel(self) -> None:
        dummy = DummyProcess("prog")
        with (
            patch("girsh.core.utils.get_processes", side_effect=[[dummy], [dummy]]),
            patch("builtins.input", side_effect=["y", "n"]),
        ):
            result: bool = utils.stop_processes("prog")
        self.assertFalse(result)

    def test_stop_processes_fail_then_kill(self) -> None:
        dummy = DummyProcess("prog")
        with (
            patch("girsh.core.utils.get_processes", side_effect=[[dummy], [dummy], []]),
            patch("builtins.input", side_effect=["y", "y"]),
        ):
            result: bool = utils.stop_processes("prog")
        self.assertTrue(result)

    def test_stop_processes_fail_then_kill_failed(self) -> None:
        dummy = DummyProcess("prog")
        with (
            patch("girsh.core.utils.get_processes", side_effect=[[dummy], [dummy], [dummy]]),
            patch("builtins.input", side_effect=["y", "y"]),
        ):
            result: bool = utils.stop_processes("prog")
        self.assertFalse(result)

    def test_stop_processes_none(self) -> None:
        result: bool = utils.stop_processes(None)
        self.assertFalse(result)


# -------------------------------
# Tests for running commands
# -------------------------------
class TestRunCommands(unittest.TestCase):
    @patch("girsh.core.utils.logger")
    def test_run_commands_success(self, mock_logger: MagicMock) -> None:
        self.assertTrue(utils.run_commands(["echo 'command 1'", "echo 'command 2'"], "test commands"))
        mock_logger.debug.assert_called_with("Command output: command 2\n, error: , error code: 0")

    @patch("girsh.core.utils.logger")
    def test_run_commands_failure(self, mock_logger: MagicMock) -> None:
        self.assertFalse(
            utils.run_commands(["sh -c 'echo failure;echo error >&2;exit 1'", "echo 'command 2'"], "test commands")
        )
        mock_logger.debug.assert_called_with("Command output: failure\n, error: error\n, error code: 1")
        mock_logger.error.assert_called_with("Command error message: error\n")

    @patch("girsh.core.utils.logger")
    @patch("girsh.core.utils.confirm_default_no", return_value=True)
    def test_run_commands_macro_confirm(self, mock_confirm: MagicMock, mock_logger: MagicMock) -> None:
        self.assertTrue(
            utils.run_commands([r"%confirm_default_no% Dummy confirm", "echo 'command 1'"], "test commands")
        )
        mock_confirm.assert_called_once_with("Dummy confirm")
        mock_logger.debug.assert_called_with("Run macro: %confirm_default_no%")

    @patch("girsh.core.utils.logger")
    @patch("subprocess.run")
    @patch("girsh.core.utils.confirm_default_no", return_value=False)
    def test_run_commands_macro_confirm_false(
        self, mock_confirm: MagicMock, mock_run: MagicMock, mock_logger: MagicMock
    ) -> None:
        self.assertFalse(
            utils.run_commands([r"%confirm_default_no% Dummy confirm", "echo 'command 1'"], "test commands")
        )
        mock_confirm.assert_called_once_with("Dummy confirm")
        mock_run.assert_not_called()

    @patch("girsh.core.utils.logger")
    @patch("girsh.core.utils.confirm_default_no", return_value=True)
    def test_run_commands_macro_confirm_no_question(self, mock_confirm: MagicMock, mock_logger: MagicMock) -> None:
        self.assertTrue(utils.run_commands([r"%confirm_default_no%", "echo 'command 1'"], "test commands"))
        mock_confirm.assert_called_once_with(None)
        mock_logger.debug.assert_called_with("Run macro: %confirm_default_no%")

    @patch("girsh.core.utils.logger")
    @patch("subprocess.run")
    @patch("girsh.core.utils.confirm_default_no")
    def test_run_commands_macro_no_close(
        self, mock_confirm: MagicMock, mock_run: MagicMock, mock_logger: MagicMock
    ) -> None:
        self.assertFalse(
            # Try incorrect macro where there is no closing %
            utils.run_commands([r"%confirm_default_no Dummy confirm", "echo 'command 1'"], "test commands")
        )
        mock_confirm.assert_not_called()
        mock_run.assert_not_called()
        mock_logger.error.assert_called_once_with(r"Unknown command macro: %confirm_default_no")

    @patch("girsh.core.utils.logger")
    @patch("subprocess.run")
    def test_run_commands_unknown_macro(self, mock_run: MagicMock, mock_logger: MagicMock) -> None:
        self.assertFalse(
            # Unknown macro
            utils.run_commands([r"%unknown_macro% Dummy confirm", "echo 'command 1'"], "test commands")
        )
        mock_run.assert_not_called()
        mock_logger.error.assert_called_once_with(r"Unknown command macro: %unknown_macro%")

    @patch("girsh.core.utils.logger")
    def test_run_commands_comment(self, mock_logger: MagicMock) -> None:
        self.assertTrue(
            utils.run_commands(
                ["#sh -c 'echo failure;echo error >&2;exit 1'", None, "echo 'command 2'"], "test commands"
            )
        )
        mock_logger.debug.assert_any_call("Skipping command: #sh -c 'echo failure;echo error >&2;exit 1'")
        mock_logger.debug.assert_any_call("Skipping command: None")
        mock_logger.debug.assert_called_with("Command output: command 2\n, error: , error code: 0")
        mock_logger.error.assert_not_called()

    @patch("girsh.core.utils.logger")
    def test_run_commands_suppress_error(self, mock_logger: MagicMock) -> None:
        self.assertTrue(
            utils.run_commands(
                ["|sh -c 'echo failure;echo error >&2;exit 1'", "echo 'command 2'"],
                "test commands",
            )
        )
        mock_logger.debug.assert_any_call("Command output: failure\n, error: error\n, error code: 1")
        mock_logger.debug.assert_called_with("Command output: command 2\n, error: , error code: 0")
        mock_logger.error.assert_not_called()

    @patch("girsh.core.utils.logger")
    @patch("girsh.core.utils.subprocess.run", side_effect=Exception("Test Exception"))
    def test_run_commands_exception(self, mock_run: MagicMock, mock_logger: MagicMock) -> None:
        self.assertFalse(
            utils.run_commands(
                ["faulty_command"],
                "test commands",
            )
        )
        mock_logger.error.assert_called_once_with("Command failed: faulty_command, error: Test Exception")

    @patch("girsh.core.utils.logger")
    @patch("girsh.core.utils.subprocess.run", return_value=MagicMock(returncode=0, stdout="42", stderr=""))
    def test_run_command_in_shell(self, mock_run: MagicMock, mock_logger: MagicMock) -> None:
        self.assertTrue(utils.run_commands(["*echo Test 42 | cut -c6-7"], "test shell"))
        mock_run.assert_called_once_with("echo Test 42 | cut -c6-7", shell=True, capture_output=True, text=True)  # noqa: S604
        mock_logger.debug.assert_called_with("Command output: 42, error: , error code: 0")


if __name__ == "__main__":
    unittest.main()



import bz2
import os
import re
import shutil
import sys
import tarfile
import zipfile
from http import HTTPStatus
from pathlib import Path
from typing import Any

import requests
from loguru import logger


def clean_downloads_folder(download_dir: Path) -> int:
    """
    Remove the downloads folder and exit.

    Args:
        download_dir (Path): The download folder path

    Returns:
        int: error code
    """
    try:
        if download_dir.exists() and download_dir.is_dir():
            logger.debug(f"Removing downloads folder: {download_dir}")
            shutil.rmtree(download_dir)
            logger.info("Downloads folder removed.")
        else:
            logger.info(f"Download folder {download_dir} doesn't exist.")
    except PermissionError:
        logger.error(f"No permission to delete the download folder {download_dir}")
        return 1
    return 0


def get_filename_from_cd(content_disposition: str | None) -> str | None:
    """
    Extract filename from Content-Disposition header.

    Args:
        content_disposition (str): The Content-Disposition header value.

    Returns:
        str: The filename if found, otherwise None.
    """
    if not content_disposition:
        return None

    # The header might look like: 'attachment; filename="archive.tar.gz"'
    logger.debug(f"Download content-disposition: {content_disposition}")
    try:
        filename_match = content_disposition.split("filename=")[1].split(";")[0]
    except IndexError:
        return None
    else:
        logger.debug(f"Matched content-disposition: {filename_match}")
        return filename_match


def download_package(download_url: str, output_dir: Path, filename: str | None = None) -> Path | None:
    """
    Downloads a file from the given URL and saves it to the specified output directory.

    Args:
        download_url (str): The URL to download the file from.
        output_dir (Path): The directory where the downloaded file will be saved.
        filename (str | None, optional): The name to save the file as. If not provided, the filename will be
                                         extracted from the Content-Disposition header or the URL.

    Returns:
        Path | None: The path to the downloaded file, or None if the download failed.
    """
    logger.debug(f"Downloading from {download_url}")
    binary_response = requests.get(download_url, stream=True, allow_redirects=True, timeout=30)
    if binary_response.ok:
        filename = filename if filename else get_filename_from_cd(binary_response.headers.get("content-disposition"))
        output_path = output_dir / (filename if filename else download_url.rsplit("/", 1)[1])
        if not output_path.is_file():
            with output_path.open("wb") as f:
                for chunk in binary_response.iter_content(chunk_size=1024):
                    if chunk:
                        f.write(chunk)
            logger.info(f"Downloaded: {output_path}")
        else:
            logger.info(f"Asset already downloaded: {output_path}")
        return output_path
    else:
        logger.error(f"Download of {download_url} failed with status: {HTTPStatus(binary_response.status_code)}")
    return None


def download_github_release(
    url: str,
    package_pattern: str,
    output_dir: Path,
    release_info: dict[Any, Any] | None = None,
    download_url: str | None = None,
) -> tuple[Path, str] | None:
    """
    Download a matching release asset from a GitHub repository.

    Args:
        url (str): GitHub API release URL.
        package_pattern (str): Regex pattern to match the desired asset.
        output_dir (Path): Directory to save the downloaded asset.
        release_info (dict[Any, Any] | None): Optionally provide release JSON to avoid duplicate API calls.
        download_url (str|None): Optional download URL template

    Returns:
        tuple[Path, str] | None: Tuple of path to the downloaded file and the release tag, or None if no match.
    """

    if release_info is None:
        try:
            response = requests.get(url, headers={"Accept": "application/vnd.github.v3+json"}, timeout=10)
            response.raise_for_status()
            release_info = response.json()
        except requests.RequestException as e:
            logger.error(f"Failed to fetch release info: {e}")
            return None

    logger.trace(f"Release info: {release_info}")
    tag = release_info.get("tag_name", "unknown")
    assets = release_info.get("assets", [])

    if download_url:
        download_url = download_url.replace("{version}", tag)
        result = download_package(download_url, output_dir)
        return (result, tag) if result else None

    logger.debug(f"Searching for package matching pattern: {package_pattern}")
    for asset in assets:
        asset_name = asset.get("name", "")
        if re.search(package_pattern, asset_name) or not package_pattern:
            download_url = asset.get("browser_download_url")
            if download_url:
                logger.debug(f"Downloading asset: {asset_name} from {download_url}")
                result = download_package(download_url, output_dir, asset_name)
                return (result, tag) if result else None
            else:
                logger.warning(f"No download URL found for asset: {asset_name}")
                return None

    logger.info("No matching asset found.")
    return None


def get_common_prefix(names: list[str]) -> str | None:
    """
    Determine the common top-level folder from a list of archive member names.

    Args:
        names (list[str]): List of member paths (as strings) from the archive.

    Returns:
        str | None: The common top-level folder if all names share one; otherwise, None.
    """
    prefixes = {name.lstrip("./").split("/", 1)[0] for name in names if name.lstrip("./")}
    return prefixes.pop() if len(prefixes) == 1 else None


def is_safe_path(base: Path, target: Path) -> bool:
    """
    Check if the resolved target path is within the resolved base directory to prevent path traversal.

    Args:
        base (Path): The intended base directory.
        target (Path): The target path to be validated.

    Returns:
        bool: True if the target is within the base directory; False otherwise.
    """
    try:
        return target.resolve().is_relative_to(base.resolve())
    except OSError:
        # General OS-related errors (e.g., invalid paths)
        return False


def extract_zip_archive(archive: zipfile.ZipFile, base: Path) -> None:
    """
    Extract members from a zip archive into the given base directory, skipping unsafe paths.

    Args:
        archive (zipfile.ZipFile): The zip archive to extract.
        base (Path): The base directory where files will be extracted.

    Returns:
        None
    """
    for name in archive.namelist():
        target = base / name
        if is_safe_path(base, target):
            archive.extract(name, base)
        else:
            logger.warning(f"Skipping unsafe member: {name}")


def extract_tar_archive(archive: tarfile.TarFile, base: Path) -> None:
    """
    Extract members from a tar archive into the given base directory, skipping unsafe paths.

    Args:
        archive (tarfile.TarFile): The tar archive to extract.
        base (Path): The base directory where files will be extracted.

    Returns:
        None
    """
    archive.extraction_filter = getattr(
        tarfile, "tar_filter", (lambda member, path: member)
    )  # Use the 'tar' filter if available, but revert to Python 3.11 behavior ('fully_trusted') if this feature is not available
    for member in archive.getmembers():
        target = base / member.name
        if is_safe_path(base, target):
            archive.extract(member, base)
        else:
            logger.warning(f"Skipping unsafe member: {member.name}")


def extract_bz2_archive(file_path: Path, extract_to: Path, package_name: str) -> None:
    """
    Extract a pure bz2 compressed file (non-tar archive) into a subfolder.

    The function decompresses a .bz2 file and writes the output into a folder
    named after package_name within extract_to.

    Args:
        file_path (Path): The path to the .bz2 file.
        extract_to (Path): The base directory where the file should be extracted.
        package_name (str): The name for the package folder.
    """
    target_folder = extract_to / package_name
    target_folder.mkdir(parents=True, exist_ok=True)
    # Remove the .bz2 suffix to obtain the original filename.
    output_file = target_folder / file_path.with_suffix("").name
    with bz2.open(file_path, "rb") as f_in, output_file.open("wb") as f_out:
        shutil.copyfileobj(f_in, f_out)
    logger.info(f"Extracted bz2 file to: {output_file}")


def extract_archive(file_path: Path, extract_to: Path, package_name: str) -> str | None:
    """
    Extract the archive based on its extension and return the common top-level folder name
    if one is detected. For pure bz2 files, extraction is done directly.

    Args:
        file_path (Path): The path to the archive.
        extract_to (Path): The base extraction directory.
        package_name (str): The package folder name used when a common folder is not found.

    Returns:
        str | None: The common top-level folder name if detected; otherwise, None.
    """
    common = None

    if file_path.suffix == ".zip":
        with zipfile.ZipFile(file_path, "r") as archive:
            names = archive.namelist()
            common = get_common_prefix(names)
            base = extract_to if common else extract_to / package_name
            if not common:
                base.mkdir(parents=True, exist_ok=True)
            extract_zip_archive(archive, base)

    elif file_path.suffixes[-2:] == [".tar", ".gz"] or file_path.suffix == ".tgz":
        with tarfile.open(file_path, "r:gz") as archive:
            members = archive.getmembers()
            names = [member.name for member in members]
            common = get_common_prefix(names)
            base = extract_to if common else extract_to / package_name
            if not common:
                base.mkdir(parents=True, exist_ok=True)
            extract_tar_archive(archive, base)

    elif file_path.suffixes[-2:] == [".tar", ".bz2"]:
        with tarfile.open(file_path, "r:bz2") as archive:
            members = archive.getmembers()
            names = [member.name for member in members]
            common = get_common_prefix(names)
            base = extract_to if common else extract_to / package_name
            if not common:
                base.mkdir(parents=True, exist_ok=True)
            extract_tar_archive(archive, base)

    elif file_path.suffix == ".bz2":
        # Handle pure bz2 compressed files.
        extract_bz2_archive(file_path, extract_to, package_name)
        return None  # Pure bz2 extraction does not yield a common folder.

    else:  # Assume that the file is a binary
        logger.info("Downloaded file is not a known archive, assume it is a binary.")
        # Move downloaded file to extract folder
        file_path.rename(extract_to / file_path.name)
        return None

    return common


def extract_package(file_path: Path, extract_to: Path, package_name: str) -> None:
    """
    Extract a compressed package file into a controlled folder structure.
    This function now delegates the extraction process to `extract_archive` to reduce complexity.

    Args:
        file_path (Path): The path to the compressed package file.
        extract_to (Path): The base directory where the package should be extracted.
        package_name (str): The name for the package folder.

    Returns:
        None
    """
    logger.info(f"Extracting {file_path} into {extract_to}")
    common = extract_archive(file_path, extract_to, package_name)

    # If a common top-level folder was detected during extraction, rename it.
    if common:
        common_folder = extract_to / common
        new_folder = extract_to / package_name
        try:
            if new_folder.exists():
                logger.debug(f"Target folder {new_folder} already exists; skipping rename.")
            else:
                common_folder.rename(new_folder)
                logger.info(f"Renamed {common_folder} to {new_folder}")
        except PermissionError as pe:
            logger.error(f"Failed to rename {common_folder} to {new_folder}: {pe}")
            return
        except OSError as oe:
            logger.error(f"Failed to rename {common_folder} to {new_folder}: {oe}")
            return
    logger.debug(f"Extraction complete. Files are in: {extract_to / package_name}")


def find_binary(extract_dir: Path, filter_pattern: str | None) -> Path | None:
    """
    Search for the binary in the extracted directory and rename it if necessary.

    Args:
        extract_dir (Path): The directory where the package was extracted.
        filter_pattern (str | None): A pattern to filter and match the binary file.

    Returns:
        Path | None: The path to the binary if found, or None if no binary is found.
    """
    if filter_pattern:
        logger.trace(f"Find {filter_pattern} in {[str(n) for n in extract_dir.rglob('*')][:3]} ...")
        for file_path in extract_dir.rglob("*"):
            if file_path.is_file() and re.search(filter_pattern, str(file_path)):
                logger.debug(f"Found matching file {file_path!s}")
                return file_path
    else:
        candidates = [p for p in extract_dir.rglob("*") if p.is_file() and os.access(p, os.X_OK)]
        if len(candidates) > 1:
            logger.info(f"Found multiple binaries, will install first one: {candidates}")
        if candidates:
            return candidates[0]
        all_files = [p for p in extract_dir.rglob("*") if p.is_file()]
        logger.warning(f"No executable binary found in {extract_dir}: {candidates}: {all_files}")
    logger.debug(f"Binary not found in {extract_dir}")
    return None


def copy_to_bin(binary_path: Path, bin_base_folder: Path, binary_name: str | None = None) -> Path:
    """
    Copy the binary file to the system or user binary directory.

    Args:
        binary_path (Path): Path to the binary file to be copied.
        bin_base_folder (Path): The binaries base folder path.
        binary_name (str | None): Optional binary name, defaults to the name from binary path

    Returns:
        Path: Path to installed binary file
    """
    try:
        bin_base_folder.mkdir(parents=True, exist_ok=True)
    except PermissionError:
        logger.error(f"Permission denied when creating bin folder '{bin_base_folder}'.")
        sys.exit(1)
    dest_path = bin_base_folder / (Path(binary_name).name if binary_name else binary_path.name)
    try:
        shutil.copy2(binary_path, dest_path)
        logger.info(f"Copied binary to: {dest_path}")
        # make executable
        dest_path.chmod(0o755)
    except PermissionError:
        logger.error(f"Permission denied to install '{dest_path}'.")
        sys.exit(1)
    except OSError as err:
        logger.error(f"Copy binary to '{dest_path}' failed with error: {err.strerror}")
        sys.exit(1)
    return dest_path


def move_to_packages(
    package_source: Path,
    package_base_folder: Path,
    bin_base_folder: Path,
    binary_path: Path,
    binary_name: str | None = None,
) -> Path:
    """
    Move the extracted package to the packages base folder and
    symlink the binary file to the system or user binary directory.

    Args:
        package_source (Path): Path to the extracted package
        package_base_folder (Path): The base package folder path.
        bin_base_folder (Path): The binaries base folder path.
        binary_path (Path): Path to the binary file to be linked.
        binary_name (str | None): Optional symlink name, defaults to the name from binary path

    Returns:
        Path: Path to installed binary file
    """
    package_path = package_base_folder / package_source.name
    dest_bin_path = bin_base_folder / (Path(binary_name).name if binary_name else binary_path.name)

    # Create package base folder if it doesn't exist
    try:
        package_base_folder.mkdir(parents=True, exist_ok=True)
    except PermissionError:
        logger.error(f"Permission denied when creating packages folder '{package_base_folder}'.")
        sys.exit(1)

    # Delete existing package folder if it exists
    if package_path.is_dir():
        logger.debug(f"Delete old package folder {package_path}")
        shutil.rmtree(package_path)

    # Move source package folder to package base folder
    try:
        package_source.rename(package_path)
        logger.info(f"Moved package to: {package_path}")
    except PermissionError:
        logger.error(f"Permission denied to move package to '{package_path}'.")
        sys.exit(1)
    except OSError as err:
        logger.error(f"Move package to '{package_path}' failed with error: {err.strerror}")
        sys.exit(1)

    # Create symlink to binary
    try:
        logger.debug(f"Create symlink from {dest_bin_path} to {package_path / binary_path.relative_to(package_source)}")
        dest_bin_path.unlink(missing_ok=True)
        dest_bin_path.symlink_to(package_path / binary_path.relative_to(package_source))
    except PermissionError:
        logger.error(f"Permission denied to create symlink to '{dest_bin_path}'.")
        sys.exit(1)
    except OSError as err:
        logger.error(f"Symlink moved package to '{dest_bin_path}' failed with error: {err.strerror}")
        sys.exit(1)
    return dest_bin_path



import shutil
from collections import defaultdict
from collections.abc import Mapping
from enum import Enum
from pathlib import Path
from typing import Any, Protocol

import requests
from loguru import logger

from girsh.core import utils
from girsh.core.files import (
    copy_to_bin,
    download_github_release,
    extract_package,
    find_binary,
    move_to_packages,
)


class RepoResult(Enum):
    # Installation status changed
    installed = 1
    uninstalled = 2
    updated = 3
    # No change requested
    skipped = 11
    dry_run_install = 12
    dry_run_uninstall = 13
    cancelled = 14
    # Installation or update failed
    update_failed = 91
    install_failed = 92
    uninstall_failed = 93
    pre_commands_failed = 94
    post_commands_failed = 95
    exception = 99


class GeneralConfig(Protocol):
    download_dir: Path
    package_base_folder: Path
    bin_base_folder: Path
    package_pattern: str


class RepositoryConfig(Protocol):
    comment: str
    version: str | None
    package_pattern: str | None
    download_url: str | None
    filter_pattern: str | None
    binary_name: str | None
    multi_file: bool
    pre_update_commands: list[str | None] | None
    post_update_commands: list[str | None] | None


def fetch_release_info(repo: str, version: str | None, reinstall: bool) -> dict[Any, Any] | None:
    """
    Fetch the release information for a given repository from GitHub.

    Args:
        repo (str): The repository identifier, in the format 'owner/repo'.
        version (str | None): The repository version
        reinstall (bool): A flag to force reinstall even if no new version is detected.

    Returns:
        dict | None: The release information in JSON format, or None if an error occurs.
    """
    releases_url = (
        f"https://api.github.com/repos/{repo}/releases/latest"
        if version is None
        else f"https://api.github.com/repos/{repo}/releases/tags/{version}"
    )
    logger.debug(f"Get release info from {releases_url}")
    try:
        response = requests.get(
            releases_url,
            headers={"Accept": "application/vnd.github.v3+json"},
            timeout=10,
        )
        response.raise_for_status()
        release_info = response.json()
        if isinstance(release_info, dict):
            return release_info
        else:
            logger.error(f"Received unexpected release info data: {release_info}")
    except requests.RequestException as e:
        logger.error(f"Error fetching release info for {repo}: {e}")
    return None


def is_new_version(installed_tag: str | None, tag: str, reinstall: bool) -> RepoResult:
    """
    Determine if the current version is different from the installed version or if a reinstall is forced.

    Args:
        installed_tag (str | None): The installed version of the repository.
        tag (str): The new version's tag name from the release info.
        reinstall (bool): A flag to force reinstall even if no new version is detected.

    Returns:
        RepoResult: Target action based on version info and reinstall flag
    """
    if installed_tag is None or reinstall:
        return RepoResult.installed
    elif installed_tag != tag:
        return RepoResult.updated
    return RepoResult.skipped


def check_repo_release(
    repo: str, target_version: str | None, current_version: str | None, reinstall: bool
) -> tuple[RepoResult, dict[Any, Any]]:
    """
    Check the release information of a repository and determine the appropriate action.

    This function fetches release information for a given repository from GitHub,
    checks if a new version is available, and determines whether to skip, reinstall,
    or proceed with the release.

    Args:
        repo (str): The name of the repository to check.
        target_version (str | None): The target version to check for. If None, the latest version is used.
        current_version (str | None): The currently installed version. If None, it assumes no version is installed.
        reinstall (bool): Whether to force reinstallation even if the current version matches the target version.

    Returns:
        RepoResult | dict[Any, Any]:
            - A dictionary containing release information if a new version is available or reinstall is forced.
            - A RepoResult enum value indicating the action taken (e.g., skipped or install_failed).
    """
    # Fetch the release information from GitHub
    release_info = fetch_release_info(repo, target_version, reinstall)
    logger.debug(
        f"Received release info url: {release_info.get('url') if isinstance(release_info, dict) else release_info}"
    )
    if not isinstance(release_info, dict):
        logger.error("No valid release info received")
        return RepoResult.install_failed, {}

    # Get the tag of the latest release
    tag = release_info.get("tag_name", "unknown")

    # Check if a new version is available or if reinstall is forced
    action = is_new_version(current_version, tag, reinstall)
    if action == RepoResult.skipped:
        logger.info(f"{repo}: No newer version than {tag} found, skipping.")
        return RepoResult.skipped, {}
    return action, release_info


def process_repository(
    repo: str,
    repo_config: RepositoryConfig,
    general: GeneralConfig,
    installed_tag: str | None,
    reinstall: bool = False,
    dry_run: bool = False,
) -> tuple[RepoResult, dict[str, str]]:
    """
    Process a single repository by checking for updates, downloading the latest release, extracting,
    optionally renaming the binary, and installing it. Updates the installed with the latest version
    information if a new version is installed.

    Args:
        repo (str): The repository identifier, in the format 'owner/repo'.
        repo_config (RepositoryConfig): The repository configuration, containing version info and other settings.
        general (GeneralConfig): General configuration settings.
        installed_tag (str | None): Tag of installed version.
        reinstall (bool): A flag to force reinstall even if the current version is up-to-date (default: False).
        dry_run (bool): A flag to simulate the process without installing the binary (default: False).

    Returns:
        tuple[RepoResult, dict[str, str]]: Result of repo installation, Repo installation data
    """
    logger.info(f"Processing '{repo}': {repo_config.comment}")

    action, release_info = check_repo_release(
        repo=repo,
        target_version=repo_config.version,
        current_version=installed_tag,
        reinstall=reinstall,
    )
    if not release_info:
        return action, {}
    tag = release_info.get("tag_name", "unknown")

    # Download the release package
    result = download_github_release(
        "",
        repo_config.package_pattern if repo_config.package_pattern is not None else general.package_pattern,
        general.download_dir,
        release_info=release_info,
        download_url=repo_config.download_url,
    )
    if not result:
        return RepoResult.install_failed, {}

    package_path, downloaded_tag = result
    repo_name = repo.split("/")[1]

    # Extract the package contents
    extract_dir = general.download_dir / "extracted"
    extract_dir.mkdir(parents=True, exist_ok=True)
    extract_package(file_path=package_path, extract_to=extract_dir, package_name=repo_name)

    # Find the binary based on the configuration
    binary_path = find_binary(extract_dir, repo_config.filter_pattern)
    if not binary_path:
        return RepoResult.install_failed, {}

    # Dry-run mode: Only check for updates without installing
    if dry_run:
        logger.info(f"Dry-run: Found {repo} version {tag}, extracted binary {binary_path}, skipping installation.")
        return RepoResult.dry_run_install, {}

    # Run pre-update commands
    if not utils.run_commands(repo_config.pre_update_commands, "repo: pre-update"):
        return RepoResult.pre_commands_failed, {}

    # Install the binary to the specified bin directory
    if repo_config.multi_file:
        install_path = move_to_packages(
            package_source=extract_dir / repo_name,
            package_base_folder=general.package_base_folder,
            bin_base_folder=general.bin_base_folder,
            binary_path=binary_path,
            binary_name=repo_config.binary_name,
        )
    else:
        install_path = copy_to_bin(
            binary_path=binary_path, bin_base_folder=general.bin_base_folder, binary_name=repo_config.binary_name
        )
    logger.info(f"{repo}: Installed {install_path.name} version {tag} to {install_path.parent}")

    # Clean up the extracted directory after installation
    shutil.rmtree(extract_dir)

    # Update the installed with the new version and binary information
    install_data = {
        "tag": tag,
        "binary": str(install_path.name),
        "path": str(install_path.parent),
        "pre_update_commands": repo_config.pre_update_commands,
    }
    if repo_config.multi_file:
        install_data["package_path"] = str(general.package_base_folder / repo_name)

    # Run post-update commands
    if not utils.run_commands(repo_config.post_update_commands, "repo: post-update"):
        return RepoResult.post_commands_failed, install_data

    return action, install_data


def process_repositories(
    repositories: Mapping[str, RepositoryConfig],
    general: GeneralConfig,
    installed: dict[str, dict],
    reinstall: list[str],
    dry_run: bool = False,
) -> tuple[dict[str, dict], dict[RepoResult, int]]:
    """
    Processes a list of repositories, installs or updates them if necessary,
    and saves the updated installation data.

    Parameters:
        repositories (Mapping[str, RepositoryConfig]): A dictionary of repository names and their corresponding Repository objects.
        general (GeneralConfig): General configuration settings.
        installed (dict[str, dict]): A dictionary containing the current installation data of the repositories.
        reinstall (list[str]): List of repositories to reinstall.
        dry_run (bool, optional): A flag to perform a dry run (no changes are actually made). Defaults to False.

    Returns:
        dict[str, dict] | None: Updated install data or None
        dict[RepoResult, int]: Summary of the installation process
    Logs:
    - Warnings and errors related to repository processing.
    - A summary of the installation or update process, including the number of repositories processed for each result type.
    """
    if not repositories:
        logger.warning("No repositories defined in config")
        return installed, {}
    # Ensure that the download folder exists
    try:
        general.download_dir.mkdir(parents=True, exist_ok=True)
    except PermissionError:
        logger.error(f"Permission denied when creating download folder '{general.download_dir}'")
        return installed, {RepoResult.exception: 1}

    summary: dict[RepoResult, int] = defaultdict(int)
    for repo, repo_config in repositories.items():
        if reinstall and repo not in reinstall:
            logger.debug(f"Skipping {repo} as it is not in the reinstall list")
            continue
        current_version = installed.get(repo, {}).get("tag")
        result, install_data = process_repository(
            repo, repo_config, general, current_version, reinstall=repo in reinstall, dry_run=dry_run
        )
        summary[result] += 1
        if result == RepoResult.updated:
            logger.success(f"{repo} updated from {current_version} to {install_data.get('tag')}")
        elif result == RepoResult.installed:
            logger.success(f"{repo} installed version {install_data.get('tag')}")
        else:
            logger.success(f"{repo}: {result.name}")
        if install_data:
            installed[repo] = install_data

    return installed, summary


def remove_binary(repo: str, info: dict[Any, Any], dry_run: bool) -> None:
    """
    Remove the binary file associated with a repository.

    Args:
        repo (str): Repository name.
        info (dict[Any, Any]): Installation info containing 'path' and 'binary'.
        dry_run (bool): Simulate removal if True.
    """
    binary_path = Path(info["path"]) / info["binary"]
    version = info.get("tag")
    if binary_path.exists():
        if dry_run:
            logger.info(f"Dry-run: {repo}: Would uninstall {binary_path.name} {version} from {binary_path}")
        else:
            try:
                binary_path.unlink()
                logger.info(f"Uninstalled {binary_path.name} {version} from {binary_path} originated from {repo}")
            except PermissionError:
                logger.error(f"Permission denied when trying to uninstall {binary_path}")


def remove_package(repo: str, info: dict[Any, Any], dry_run: bool) -> None:
    """
    Remove the package directory if it exists.

    Args:
        repo (str): Repository name.
        info (dict[Any, Any]): Installation info that may contain 'package_path'.
        dry_run (bool): Simulate removal if True.
    """
    package_path_str = info.get("package_path")
    if package_path_str:
        package_path = Path(package_path_str)
        if package_path.is_dir():
            if dry_run:
                logger.info(f"Dry-run: {repo}: Would delete package folder {package_path}")
            else:
                try:
                    shutil.rmtree(package_path)
                    logger.info(f"Deleted package folder {package_path}")
                except PermissionError:
                    logger.error(f"Permission denied when trying to delete {package_path}")


def uninstall_binary(repo: str, install_data: dict, dry_run: bool) -> RepoResult:
    """
    Uninstalls a binary file associated with a given repository.

    Args:
        repo (str): The name of the repository associated with the binary.
        install_data (dict): A dictionary containing install information about the binary.
        dry_run (bool): If True, simulates the uninstallation process without making any changes.

    Returns:
        RepoResult: An enumeration indicating the result of the uninstallation process.
            Possible values:
                - RepoResult.dry_run_uninstall: Indicates a successful dry-run.
                - RepoResult.uninstalled: Indicates the binary was successfully uninstalled.
                - RepoResult.uninstall_failed: Indicates the uninstallation failed due to permission issues.
    """

    binary_path = Path(install_data["path"], install_data["binary"])
    if binary_path.exists():
        if dry_run:
            logger.success(f"Dry-run: {repo}: Would uninstall {binary_path.name} from {binary_path}.")
            # summary[RepoResult.dry_run_uninstall] += 1
            return RepoResult.dry_run_uninstall
        else:
            utils.run_commands(install_data.get("pre_update_commands"), "repo: pre-uninstall")
            try:
                binary_path.unlink()
                logger.success(f"Uninstalled {binary_path.name} from {binary_path} originated from {repo}")
                # uninstall_repos.append(repo)
                # summary[RepoResult.uninstalled] += 1
            except PermissionError:
                logger.error(f"Permission denied when trying to uninstall {binary_path}")
                # summary[RepoResult.uninstall_failed] += 1
                return RepoResult.uninstall_failed
            # else:
            #     return RepoResult.uninstalled
    return RepoResult.uninstalled


def uninstall(repositories: list[str], installed: dict[Any, Any], dry_run: bool = False) -> dict[RepoResult, int]:
    """
    Uninstall binaries which are not present in the repositories config anymore.
    If repositories is empty, all binaries will be uninstalled.
    This function will also remove the package folder if it exists.

    Args:
        repositories (list[str]): A list of repository names.
        installed (dict[Any, Any]): installed tracking installed versions and binaries.
        dry_run (bool): A flag to simulate the process without removing the binary (default: False).
    """
    if not installed:
        logger.log("INFO" if dry_run else "SUCCESS", "No binaries installed.")
        return {}
    if dry_run:
        logger.info("Check for repos to uninstall")
    elif repositories:
        logger.info("Uninstall binaries no longer in config.")
    elif utils.confirm_default_no("Are you sure that all programs should be uninstalled?"):
        logger.info("Uninstall all binaries.")
    else:
        logger.info("Uninstall cancelled.")
        return {RepoResult.cancelled: 1}
    uninstall_repos = []
    summary: dict[RepoResult, int] = defaultdict(int)
    for repo, install_data in installed.items():
        if repo not in repositories:
            uninstall_result = uninstall_binary(repo=repo, install_data=install_data, dry_run=dry_run)
            if uninstall_result == RepoResult.uninstalled:
                uninstall_repos.append(repo)
            summary[uninstall_result] += 1
    for repo in uninstall_repos:
        del installed[repo]
    return summary



import io
import tarfile
import tempfile
import unittest
from pathlib import Path
from unittest.mock import patch

from loguru import logger

# Import the module using the new submodule path
from girsh.core import files

# -------------------------------
# Tests for extract_package
# -------------------------------


class ExtractPackageTest(unittest.TestCase):
    def setUp(self) -> None:
        logger.remove()  # Remove all loguru handlers to silence logging output
        self.temp_dir: tempfile.TemporaryDirectory = tempfile.TemporaryDirectory()
        self.temp_path: Path = Path(self.temp_dir.name)
        self.extract_to = self.temp_path / "extract"
        self.extract_to.mkdir()
        self.package_name = "test_package"
        self.package_path = self.extract_to / "package"
        self.package_path.mkdir()

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    def test_extract_package_common(self) -> None:
        pkg_dir = self.extract_to / self.package_name
        self.assertFalse(pkg_dir.is_dir())
        with patch("girsh.core.files.extract_archive", return_value=str(self.package_path)):
            files.extract_package(self.package_path, self.extract_to, self.package_name)

        self.assertTrue(pkg_dir.is_dir())

    def test_extract_package_rename_permission_error(self) -> None:
        pkg_dir = self.extract_to / self.package_name
        self.assertFalse(pkg_dir.is_dir())
        with (
            patch("girsh.core.files.extract_archive", return_value=str(self.package_path)),
            patch("pathlib.Path.rename", side_effect=PermissionError("permission_error")),
            patch.object(logger, "error") as mock_logger_error,
        ):
            files.extract_package(self.package_path, self.extract_to, self.package_name)
            mock_logger_error.assert_called_once_with(
                f"Failed to rename {self.package_path} to {pkg_dir}: permission_error"
            )
        self.assertFalse(pkg_dir.is_dir())

    def test_extract_package_os_error(self) -> None:
        pkg_dir = self.extract_to / self.package_name
        self.assertFalse(pkg_dir.is_dir())
        with (
            patch("girsh.core.files.extract_archive", return_value=str(self.package_path)),
            patch("pathlib.Path.rename", side_effect=OSError("os_error")),
            patch.object(logger, "error") as mock_logger_error,
        ):
            files.extract_package(self.package_path, self.extract_to, self.package_name)
            mock_logger_error.assert_called_once_with(f"Failed to rename {self.package_path} to {pkg_dir}: os_error")
        self.assertFalse(pkg_dir.is_dir())

    def test_extract_package_no_common(self) -> None:
        zip_file_path = self.temp_path / "package.zip"
        package_name = "my_package"
        with (
            patch("girsh.core.files.extract_archive", return_value=None),
            patch.object(logger, "debug") as mock_logger,
        ):
            files.extract_package(zip_file_path, self.extract_to, package_name)
        mock_logger.assert_called_with(
            f"Extraction complete. Files are in: {self.temp_path / 'extract' / 'my_package'}"
        )

    def test_extract_package_new_folder_exists(self) -> None:
        zip_file_path = self.temp_path / "package.zip"
        package_name = "my_package"
        pkg_dir = self.extract_to / package_name
        pkg_dir.mkdir()
        with (
            patch("girsh.core.files.extract_archive", return_value="common"),
            patch.object(logger, "debug") as mock_logger,
        ):
            files.extract_package(zip_file_path, self.extract_to, package_name)
        self.assertTrue(pkg_dir.is_dir())
        mock_logger.assert_any_call(f"Target folder {pkg_dir} already exists; skipping rename.")

    def test_extract_package_tar(self) -> None:
        tar_file_path = self.temp_path / "package.tar.gz"
        package_name = "my_package"
        with tarfile.open(tar_file_path, "w:gz") as tar:
            info = tarfile.TarInfo(name="common/file.txt")
            content = b"tar content"
            info.size = len(content)
            tar.addfile(info, fileobj=io.BytesIO(content))

        files.extract_package(tar_file_path, self.extract_to, package_name)
        pkg_dir = self.extract_to / package_name
        self.assertTrue(pkg_dir.is_dir())
        self.assertTrue((pkg_dir / "file.txt").exists())

    def test_extract_package_unsupported(self) -> None:
        txt_file = self.temp_path / "file.txt"
        txt_file.write_text("not an archive")
        files.extract_package(txt_file, self.extract_to, "pkg")
        self.assertFalse((self.extract_to / "pkg").exists())


if __name__ == "__main__":
    unittest.main()



import sys
from pathlib import Path
from typing import Any

import yaml
from loguru import logger

from girsh.core.config import Repository


def load_installed(installed_file: Path) -> dict[str, dict]:
    """
    Load the version installed from a YAML file.

    Args:
        installed_file (Path): Path to the installed file.

    Returns:
        dict[str, str]: Dict of installed software details.
    """
    # Ensure that the folder for the installed file exists, so that updates can be saved later
    try:
        installed_file.parent.mkdir(parents=True, exist_ok=True)
    except PermissionError:
        logger.error(f"Permission denied when creating installed folder '{installed_file.parent}'.")
        sys.exit(1)
    try:
        if installed_file.exists():
            with installed_file.open("r") as f:
                data = yaml.safe_load(f)
                return data if isinstance(data, dict) else {}
        else:
            logger.info(f"Installed packages file '{installed_file}' not found. Proceeding with empty installed.")
    except PermissionError:
        logger.error(
            f"Warning: Permission denied when reading installed file '{installed_file}'. Proceeding with empty installed."
        )
    return {}


def save_installed(installed_file: Path, data: dict[Any, Any]) -> int:
    """
    Save the version installed to a YAML file.

    Args:
        installed_file (Path): Path to the installed file.
        data (dict[Any, Any]): Version installed data to save.

    Returns:
        int: error code
    """
    try:
        with installed_file.open("w") as f:
            yaml.safe_dump(data, f)
    except PermissionError:
        logger.error(f"Permission denied when writing to installed file '{installed_file}'.")
        return 1
    except FileNotFoundError:
        logger.error(f"installed file '{installed_file}' not found and cannot be created.")
        return 1
    return 0


def get_comment(repo: str, repositories: dict[str, Repository]) -> str:
    """
    Get the comment from the repository config.
    If the repository is not found, return a default message.

    Args
        repo (str): Name of the repository

    Returns:
        str: Comment from repository config
    """
    if repo in repositories:
        return repositories[repo].comment
    logger.debug(f"The repository {repo} is not present in current config.")
    return "Not found in config."


def show_installed(data: dict[Any, Any], repositories: dict[str, Repository]) -> int:
    """
    Display installed binaries in a table format with auto-fitted column widths.

    Args:
        data (dict[Any, Any]): Dictionary containing installed binaries with metadata.

    Returns:
        int: Error code (0 for success).
    """
    if not data:
        logger.success("Nothing installed via `girsh`")
        return 0

    # Table headers
    headers = ["Repository", "Comment", "Binary", "Tag"]

    # Gather row data
    rows = [(repo, get_comment(repo, repositories), status["binary"], status["tag"]) for repo, status in data.items()]

    # Calculate max column widths
    col_widths = [max(len(str(row[i])) for row in [headers, *rows]) for i in range(len(headers))]

    # Format line separator
    separator = "+".join("-" * (w + 2) for w in col_widths)
    separator = f"+{separator}+"

    # Print table
    logger.success("Currently installed binaries:")
    logger.success(separator)
    logger.success("| " + " | ".join(f"{headers[i]:<{col_widths[i]}}" for i in range(len(headers))) + " |")
    logger.success(separator)

    for row in rows:
        logger.success("| " + " | ".join(f"{row[i]:<{col_widths[i]}}" for i in range(len(row))) + " |")

    logger.success(separator)

    return 0



import bz2
import io
import shutil
import tarfile
import tempfile
import unittest
import zipfile
from pathlib import Path
from typing import Any
from unittest.mock import patch

from loguru import logger

# Import the module using the new submodule path
from girsh.core import files


# A base test class to set up temporary directories and dummy config.
class BaseTest(unittest.TestCase):
    def setUp(self) -> None:
        logger.remove()  # Remove all loguru handlers to silence logging output
        self.temp_dir: tempfile.TemporaryDirectory = tempfile.TemporaryDirectory()
        self.temp_path: Path = Path(self.temp_dir.name)
        self.extract_to = self.temp_path / "extract"
        self.extract_to.mkdir()
        self.package_name = "test_package"

    def tearDown(self) -> None:
        self.temp_dir.cleanup()


def fake_is_safe_path(base: Path, target: Path) -> bool:
    return "unsafe.txt" not in str(target)


# -------------------------------
# Tests for extract_archive
# -------------------------------


class ExtractArchiveTest(BaseTest):
    def test_extract_zip_no_common(self) -> None:
        zip_file = self.temp_path / "test.zip"
        with zipfile.ZipFile(zip_file, "w") as zf:
            zf.writestr("dummy", "dummy content")
        with patch("girsh.core.files.get_common_prefix", return_value=None):
            result = files.extract_archive(zip_file, self.extract_to, self.package_name)
        self.assertIsNone(result)

    def test_extract_gz_no_common(self) -> None:
        gz_file = self.temp_path / "test.tar.gz"
        with tarfile.open(gz_file, "w:gz") as tar:
            info = tarfile.TarInfo(name="safe.txt")
            content = b"safe content"
            info.size = len(content)
            tar.addfile(info, fileobj=io.BytesIO(content))
        with patch("girsh.core.files.get_common_prefix", return_value=None):
            result = files.extract_archive(gz_file, self.extract_to, self.package_name)
        self.assertIsNone(result)

    def test_extract_zip(self) -> None:
        zip_file = self.temp_path / "test.zip"
        with zipfile.ZipFile(zip_file, "w") as zf:
            zf.writestr("dummy", "dummy content")
        with (
            patch("girsh.core.files.get_common_prefix", return_value="common"),
            patch("girsh.core.files.extract_zip_archive") as mock_extract_zip_archive,
        ):
            result = files.extract_archive(zip_file, self.extract_to, self.package_name)
            mock_extract_zip_archive.assert_called_once()
        self.assertEqual(result, "common")

    def test_extract_tar_bz2(self) -> None:
        bz2_file = self.temp_path / "test.tar.bz2"
        with tarfile.open(bz2_file, "w:bz2") as tar:
            info = tarfile.TarInfo(name="safe.txt")
            content = b"safe content"
            info.size = len(content)
            tar.addfile(info, fileobj=io.BytesIO(content))
        with (
            patch("girsh.core.files.get_common_prefix", return_value="common"),
            patch("girsh.core.files.extract_tar_archive") as mock_extract_tar_archive,
        ):
            result = files.extract_archive(bz2_file, self.extract_to, self.package_name)
            mock_extract_tar_archive.assert_called_once()
        self.assertEqual(result, "common")

    def test_extract_tar_bz2_no_common(self) -> None:
        bz2_file = self.temp_path / "test.tar.bz2"
        with tarfile.open(bz2_file, "w:bz2") as tar:
            info = tarfile.TarInfo(name="safe.txt")
            content = b"safe content"
            info.size = len(content)
            tar.addfile(info, fileobj=io.BytesIO(content))
        with patch("girsh.core.files.get_common_prefix", return_value=None):
            result = files.extract_archive(bz2_file, self.extract_to, self.package_name)
        self.assertIsNone(result)

    def test_extract_bz2(self) -> None:
        bz2_file = self.temp_path / "test.bz2"
        with tarfile.open(bz2_file, "w:bz2") as tar:
            info = tarfile.TarInfo(name="safe.txt")
            content = b"safe content"
            info.size = len(content)
            tar.addfile(info, fileobj=io.BytesIO(content))
        with (
            patch("girsh.core.files.extract_bz2_archive") as mock_extract_bz2_archive,
        ):
            result = files.extract_archive(bz2_file, self.extract_to, self.package_name)
            mock_extract_bz2_archive.assert_called_once()
        self.assertIsNone(result)

    def test_extract_is_binary(self) -> None:
        bin_file = self.temp_path / "test_bin"
        with bin_file.open("wb") as f:
            f.write(b"binary content")
        result = files.extract_archive(bin_file, self.extract_to, self.package_name)
        self.assertIsNone(result)
        self.assertTrue((self.extract_to / "test_bin").exists())


# -------------------------------
# Tests for extract_zip_archive
# -------------------------------


class ExtractZipArchiveTest(BaseTest):
    def test_extract_zip_archive_safe(self) -> None:
        zip_path = self.temp_path / "test.zip"
        with zipfile.ZipFile(zip_path, "w") as zf:
            zf.writestr("safe.txt", "safe content")
            zf.writestr("unsafe.txt", "unsafe content")
        with (
            zipfile.ZipFile(zip_path, "r") as archive,
            patch("girsh.core.files.is_safe_path", side_effect=fake_is_safe_path),
        ):
            extracted = []
            original_extract = archive.extract

            def fake_extract(name: str, base_dir: Path) -> Any:
                extracted.append(name)
                return original_extract(name, base_dir)

            with patch.object(archive, "extract", side_effect=fake_extract):
                files.extract_zip_archive(archive, self.extract_to)
            self.assertIn("safe.txt", extracted)
            self.assertNotIn("unsafe.txt", extracted)


# -------------------------------
# Tests for extract_tar_archive
# -------------------------------


class ExtractTarArchiveTest(BaseTest):
    def test_extract_tar_archive_safe(self) -> None:
        temp_dir = Path(tempfile.mkdtemp())
        try:
            tar_path = temp_dir / "test.tar.gz"
            with tarfile.open(tar_path, "w:gz") as tar:
                info = tarfile.TarInfo(name="safe.txt")
                content = b"safe content"
                info.size = len(content)
                tar.addfile(info, fileobj=io.BytesIO(content))
                info2 = tarfile.TarInfo(name="unsafe.txt")
                content2 = b"unsafe content"
                info2.size = len(content2)
                tar.addfile(info2, fileobj=io.BytesIO(content2))
            with (
                tarfile.open(tar_path, "r:gz") as archive,
                patch("girsh.core.files.is_safe_path", side_effect=fake_is_safe_path),
            ):
                extract_dir = temp_dir / "extract"
                extract_dir.mkdir()
                extracted = []
                original_extract = archive.extract

                def fake_extract(member: tarfile.TarInfo, base_dir: Path) -> Any:
                    extracted.append(member.name)
                    return original_extract(member, base_dir)

                with patch.object(archive, "extract", side_effect=fake_extract):
                    files.extract_tar_archive(archive, extract_dir)
                self.assertIn("safe.txt", extracted)
                self.assertNotIn("unsafe.txt", extracted)
        finally:
            shutil.rmtree(temp_dir)


# -------------------------------
# Tests for extract_bz2_archive
# -------------------------------


class ExtractBZ2ArchiveTest(BaseTest):
    def test_extract_bz2_archive_success(self) -> None:
        bz2_file = self.temp_path / "test.bz2"
        content = b"test content"
        with bz2.BZ2File(bz2_file, "wb") as f:
            f.write(content)

        files.extract_bz2_archive(bz2_file, self.extract_to, self.package_name)

        extracted_file = self.extract_to / self.package_name / "test"
        self.assertTrue(extracted_file.exists())
        self.assertEqual(extracted_file.read_bytes(), content)

    def test_extract_bz2_archive_no_permission(self) -> None:
        bz2_file = self.temp_path / "test.bz2"
        content = b"test content"
        with bz2_file.open("wb") as f:
            f.write(content)

        with patch("pathlib.Path.mkdir", side_effect=PermissionError), self.assertRaises(PermissionError):
            files.extract_bz2_archive(bz2_file, self.extract_to, self.package_name)

    def test_extract_bz2_archive_os_error(self) -> None:
        bz2_file = self.temp_path / "test.bz2"
        content = b"test content"
        with bz2_file.open("wb") as f:
            f.write(content)

        with patch("shutil.copyfileobj", side_effect=OSError), self.assertRaises(OSError):
            files.extract_bz2_archive(bz2_file, self.extract_to, self.package_name)


if __name__ == "__main__":
    unittest.main()



import tempfile
import unittest
from collections.abc import Iterator
from http import HTTPStatus
from pathlib import Path
from unittest.mock import Mock, patch

from loguru import logger

# Import the module using the new submodule path
from girsh.core import files

# -------------------------------
# FakeResponse for download_package tests
# -------------------------------


class FakeResponse:
    def __init__(self, headers: dict[str, str], content_chunks: list[bytes], status_code: int = 200) -> None:
        self.headers = headers
        self.status_code = status_code
        self._content_chunks = content_chunks
        self.ok = self.status_code < 400

    def iter_content(self, chunk_size: int) -> Iterator[bytes]:
        yield from self._content_chunks


def fake_requests_get_success(url: str, stream: bool, allow_redirects: bool, timeout: int) -> FakeResponse:
    headers = {"content-disposition": 'attachment; filename="testfile.txt"'}
    return FakeResponse(headers, [b"data"])


# -------------------------------
# Tests for download_package
# -------------------------------


class DownloadPackageTest(unittest.TestCase):
    def setUp(self) -> None:
        logger.remove()  # Remove all loguru handlers to silence logging output
        self.temp_dir: tempfile.TemporaryDirectory = tempfile.TemporaryDirectory()
        self.output_dir: Path = Path(self.temp_dir.name)

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    def test_download_package_request_unauthorized(self) -> None:
        def fake_get(url: str, stream: bool, allow_redirects: bool, timeout: int) -> FakeResponse:
            return FakeResponse({}, [b"chunk"], status_code=401)

        download_url = "http://example.com/file"
        with patch("requests.get", side_effect=fake_get), patch.object(logger, "error") as mock_logger:
            files.download_package(download_url, self.output_dir)
            mock_logger.assert_called_once_with(f"Download of {download_url} failed with status: {HTTPStatus(401)}")

    def test_download_package_with_filename(self) -> None:
        with patch("requests.get", side_effect=fake_requests_get_success):
            result = files.download_package("http://example.com/file", self.output_dir, "explicit.txt")
        self.assertIsNotNone(result)
        if result is not None:
            self.assertEqual(result.name, "explicit.txt")
            self.assertTrue(result.exists())
            self.assertEqual(result.read_bytes(), b"data")

    def test_download_package_without_filename(self) -> None:
        with patch("requests.get", side_effect=fake_requests_get_success):
            result = files.download_package("http://example.com/file", self.output_dir)
        self.assertIsNotNone(result)
        if result is not None:
            self.assertEqual(result.name, '"testfile.txt"')
            self.assertTrue(result.exists())
            self.assertEqual(result.read_bytes(), b"data")

    def test_download_package_no_header(self) -> None:
        def fake_get(url: str, stream: bool, allow_redirects: bool, timeout: int) -> FakeResponse:
            return FakeResponse({}, [b"chunk"])

        with patch("requests.get", side_effect=fake_get):
            result = files.download_package("http://example.com/file", self.output_dir)
        self.assertIsNotNone(result)
        if result is not None:
            self.assertEqual(result.name, "file")
            self.assertTrue(result.exists())
            self.assertEqual(result.read_bytes(), b"chunk")

    def test_download_package_already_downloaded(self) -> None:
        existing = self.output_dir / '"existing.txt"'
        existing.write_bytes(b"existing")

        def fake_get(url: str, stream: bool, allow_redirects: bool, timeout: int) -> FakeResponse:
            headers = {"content-disposition": 'attachment; filename="existing.txt"'}
            return FakeResponse(headers, [b"new data"])

        with patch("requests.get", side_effect=fake_get):
            result = files.download_package("http://example.com/existing.txt", self.output_dir)
        self.assertIsNotNone(result)
        if result is not None:
            self.assertEqual(result.read_bytes(), b"existing")

    def test_download_package_empty_chunk(self) -> None:
        with patch("requests.get") as mock_get:
            mock_response = Mock()
            mock_response.iter_content.return_value = [b"", b"data"]
            mock_response.headers = {"content-disposition": 'attachment; filename="testfile.txt"'}
            mock_get.return_value = mock_response

            result = files.download_package("http://example.com/file", self.output_dir)
            self.assertIsNotNone(result)
            if result is not None:
                self.assertEqual(result.name, '"testfile.txt"')
                self.assertTrue(result.exists())
                self.assertEqual(result.read_bytes(), b"data")


if __name__ == "__main__":
    unittest.main()



import shlex
import subprocess

import psutil
from loguru import logger


def confirm_default_no(question: str | None) -> bool:
    """
    Prompts the user with a yes/no question and returns True if the user explicitly answers "y" (yes).
    Args:
        question (str): The question to display to the user.
    Returns:
        bool: True if the user answers "y", otherwise False.
    """
    answer = input(f"{question if question else 'Confirm?'} (y/N): ").strip().lower()
    return answer == "y"


def confirm_default_yes(question: str | None) -> bool:
    """
    Prompts the user with a yes/no question and returns True if the user doesn't explicitly answers "n" (no).
    Args:
        question (str): The question to display to the user.
    Returns:
        bool: True if the user doesn't answers "n", otherwise False.
    """
    answer = input(f"{question if question else 'Confirm?'} (Y/n): ").strip().lower()
    return answer != "n"


def get_processes(program_name: str) -> list[psutil.Process]:
    """Search for matches in the process name

    Args:
        program_name (str): Name of the program

    Returns:
        list[int]: List of processes
    """
    return [proc for proc in psutil.process_iter(["pid", "name", "username"]) if proc.info["name"] == program_name]


def stop_processes(program_name: str | None) -> bool:
    """
    Terminate all running processes for of given program

    Args:
        program_name (str): Name of the program

    Returns:
        bool: True if all processes have been terminated.
    """
    if program_name is None:
        logger.error("To stop a process, the program name must be specified!")
        return False
    processes = get_processes(program_name)
    if not processes:
        return True
    if not confirm_default_no(f"Terminate all running processes of '{program_name}'?"):
        logger.warning(f"Can't install or update {program_name} due to running processes.")
        return False
    for process in processes:
        process.terminate()
    # Check if all have been stopped
    processes = get_processes(program_name)
    if not processes:
        logger.warning(f"Successfully stopped all processes of {program_name}")
        return True
    logger.warning(f"Failed to stop all running processes of {program_name}")
    logger.debug(f"Failed to terminate processes: {[proc.info for proc in processes]}")
    if not confirm_default_no(f"Kill all running processes of '{program_name}'?"):
        logger.warning(f"Can't install or update {program_name} due to running processes.")
        return False
    for process in processes:
        process.kill()
    processes = get_processes(program_name)
    if not processes:
        logger.info(f"Successfully killed all running processes of {program_name}")
        return True
    logger.error(f"Failed to kill processes: {[proc.info for proc in processes]}")
    return False


def run_macro(command: str) -> bool:
    """
    Executes a command macro if it is found in the command string.
    The command macros are defined in the `command_macros` dictionary.

    Args:
        command (str): The command string.

    Returns:
        bool | None: True if macro was found and successfully executed,
                     False if macro was not found or failed.
    """
    command_macros = {
        r"%confirm_default_yes%": confirm_default_yes,
        r"%confirm_default_no%": confirm_default_no,
        r"%stop_processes%": stop_processes,
    }
    if " " in command:
        macro, macro_args = command.split(" ", 1)
    else:
        macro = command
        macro_args = None

    if macro in command_macros:
        logger.debug(f"Run macro: {macro}")
        return command_macros[macro](macro_args)

    logger.error(f"Unknown command macro: {macro}")
    return False


def run_commands(command_list: list[str | None] | None, info: str) -> bool:
    """
    Executes a list of shell commands and logs their output.

    Args:
        command_list (list[str|None] | None): A list of shell commands to execute. If None, no commands are run.
        info (str): A descriptive string used for logging purposes.

    Returns:
        bool: True if all commands execute successfully or if the command list is None.
              False if any command fails.
    """
    # Define dict of command macros
    # The command macro functions must take one optional string argument and return True if successfully executed.
    if command_list:
        logger.info(f"Running commands for {info}")
        for command in command_list:
            if command is None or command.startswith("#"):
                logger.debug(f"Skipping command: {command}")
                continue
            logger.debug(f"Running command: {command}")
            if command.startswith("%"):
                return run_macro(command)
            # Command was no macro, so try to execute it
            if command.startswith("|"):
                fail_on_error = False
                command = command[1:]
            else:  # default is to fail on error
                fail_on_error = True
            try:
                if command.startswith("*"):
                    # If the command starts with "*", it is executed in a shell
                    command = command[1:]
                    cmd_result = subprocess.run(command, shell=True, capture_output=True, text=True)  # noqa: S602
                else:
                    cmd_result = subprocess.run(shlex.split(command), capture_output=True, text=True)  # noqa: S603
            except Exception as e:
                logger.error(f"Command failed: {command}, error: {e}")
                return False

            logger.debug(
                f"Command output: {cmd_result.stdout}, error: {cmd_result.stderr}, error code: {cmd_result.returncode}"
            )
            if fail_on_error and cmd_result.returncode != 0:
                logger.error(f"Commands for {info} failed: {command}, error code: {cmd_result.returncode}")
                logger.error(f"Command error message: {cmd_result.stderr}")
                return False
    return True



import os
import sys

from loguru import logger

from girsh.core.config import edit_config, get_arguments, load_yaml_config
from girsh.core.files import clean_downloads_folder
from girsh.core.installed import load_installed, save_installed, show_installed
from girsh.core.repos import RepoResult, process_repositories, uninstall

GITHUB_API_RELEASES = "https://api.github.com/repos/{repo}/releases"

logger.remove(0)
logger.add(sys.stdout, colorize=True, format="<level>{message}</level>", level="SUCCESS")


def elevate_privileges() -> None:
    """
    Elevate the privileges of the current process to root using sudo.
    If the current process is not running as the root user, this function
    will re-run the script with elevated privileges by invoking sudo.
    Raises:
        OSError: If there is an error executing the sudo command.
    """

    if os.geteuid() != 0:
        print("Re-running with elevated privileges...")
        command = ["sudo", sys.executable, *sys.argv]
        os.execvp("/usr/bin/sudo", command)  # Replace current process with sudo call  # noqa: S606


def show_summary(install_summary: dict[RepoResult, int], uninstall_summary: dict[RepoResult, int]) -> None:
    """
    Display a summary of the repository processing results.

    Args:
        install_summary (dict[RepoResult, int]): A dictionary where the keys are
            RepoResult instances and the values are counts of occurrences.
        uninstall_summary (dict[RepoResult, int]): A dictionary where the keys are
            RepoResult instances and the values are counts of occurrences.
    """
    logger.success("===============================")
    logger.success("Summary:")
    for result, count in install_summary.items():
        logger.success(f"  {result.name}: {count}")
    for result, count in uninstall_summary.items():
        logger.success(f"  {result.name}: {count}")


def set_logger_level(verbosity: int) -> None:
    """
    Set the logger level based on the verbosity argument.

    Args:
        verbosity (int): The verbosity level (0-3).
    """
    if verbosity == 0:
        return
    levels = ["INFO", "DEBUG", "TRACE"]
    logger.remove()
    logger.add(
        sys.stdout,
        colorize=True,
        format="<level>{message}</level>",
        level=levels[verbosity - 1 if verbosity < 4 else 2],
    )


def main() -> int:
    """
    The main entry point for the application.

    This function processes command-line arguments, manages configurations,
    and orchestrates the execution of various operations such as editing
    configurations, cleaning the downloads folder, showing installed items,
    uninstalling repositories, and processing repositories for installation.

    Returns:
        int: Exit code indicating the result of the operation.
             - 0: Success
             - 1: Uninstall operation failed
             - 3: No repositories installed
             - Other non-zero values indicate errors or specific conditions
    """

    args = get_arguments()

    if args.system:
        elevate_privileges()

    if args.edit:
        return edit_config(args.config)

    set_logger_level(args.verbose)

    general, repositories = load_yaml_config(args.config)
    logger.debug(f"General config: {general}")
    logger.debug(f"Repositories config: {repositories}")

    if args.clean:
        return clean_downloads_folder(general.download_dir)

    installed: dict[str, dict] = load_installed(general.installed_file)
    logger.debug(f"Current installed: {installed}")

    if args.show:
        return show_installed(installed, repositories)

    if args.uninstall_all:
        uninstall_summary = uninstall(repositories=[], installed=installed, dry_run=args.dry_run)
    else:
        uninstall_summary = uninstall(
            repositories=list(repositories), installed=installed, dry_run=args.dry_run if args.uninstall else True
        )
    if args.uninstall or args.uninstall_all:
        save_installed(general.installed_file, installed)
        return 1 if RepoResult.uninstall_failed in uninstall_summary else 0

    # Determine binaries to reinstall using a ternary operator
    reinstall_repos = (
        [repo for repo in installed if installed[repo].get("binary") in args.reinstall] if args.reinstall else []
    )

    installed, summary = process_repositories(
        repositories, general, installed, reinstall=reinstall_repos, dry_run=args.dry_run
    )

    show_summary(summary, uninstall_summary)

    if repositories and not installed:
        # Nothing installed
        return 3
    has_errors = summary.get(RepoResult.install_failed, 0) > 0
    # Update the installed with the new version
    return save_installed(general.installed_file, installed) + has_errors


if __name__ == "__main__":
    sys.exit(main())



import shutil
import tempfile
import unittest
from pathlib import Path
from unittest.mock import patch

from loguru import logger

# Import the module using the new submodule path
from girsh.core import files

# -------------------------------
# Tests for clean_downloads_folder
# -------------------------------


class CleanDownloadsFolderTest(unittest.TestCase):
    def setUp(self) -> None:
        logger.remove()  # Remove all loguru handlers to silence logging output
        self.temp_dir: tempfile.TemporaryDirectory = tempfile.TemporaryDirectory()
        self.download_dir: Path = Path(self.temp_dir.name)
        self.dummy_file = self.download_dir / "dummy_file.txt"
        self.dummy_file.touch()  # Create a dummy file in the download folder

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    def test_clean_downloads_folder_success(self) -> None:
        self.assertTrue(self.download_dir.exists())
        ret = files.clean_downloads_folder(download_dir=self.download_dir)
        self.assertEqual(ret, 0)
        self.assertFalse(self.download_dir.exists())
        self.assertFalse(self.dummy_file.exists())

    def test_clean_downloads_folder_no_folder(self) -> None:
        shutil.rmtree(self.download_dir)
        with patch.object(logger, "info") as mock_logger_info:
            ret = files.clean_downloads_folder(download_dir=self.download_dir)
            mock_logger_info.assert_called_once_with(f"Download folder {self.download_dir} doesn't exist.")
        self.assertEqual(ret, 0)

    def test_clean_downloads_folder_permission_error(self) -> None:
        with patch("shutil.rmtree", side_effect=PermissionError), patch.object(logger, "error") as mock_logger_error:
            ret = files.clean_downloads_folder(download_dir=self.download_dir)
            mock_logger_error.assert_called_once_with(
                f"No permission to delete the download folder {self.download_dir}"
            )
        self.assertEqual(ret, 1)


if __name__ == "__main__":
    unittest.main()



import os
import tempfile
import unittest
from pathlib import Path
from typing import Any
from unittest.mock import patch

from loguru import logger

# Import the module using the new submodule path
from girsh.core import files


class SubstringMatcher:
    def __init__(self, containing: str) -> None:
        self.containing: str = containing.lower()

    def __eq__(self, other: Any) -> Any:
        return other.lower().find(self.containing) > -1

    def __unicode__(self) -> str:
        return f'a string containing "{self.containing}"'

    def __str__(self) -> str:
        return self.__unicode__().encode("utf-8").decode("utf-8")

    __repr__ = __unicode__


# -------------------------------
# Tests for move_to_packages
# -------------------------------


class MoveToPackagesTest(unittest.TestCase):
    def setUp(self) -> None:
        logger.remove()  # Remove all loguru handlers to silence logging output
        self.temp_dir: tempfile.TemporaryDirectory = tempfile.TemporaryDirectory()
        self.temp_path: Path = Path(self.temp_dir.name)
        self.bin_base_folder = self.temp_path / "bin"
        self.bin_base_folder.mkdir()
        self.package_base_folder = self.temp_path / "packages"
        self.package_base_folder.mkdir()
        self.package_source = self.temp_path / "package_source"
        self.package_source.mkdir()
        self.binary_rel = "bin/executable"
        self.binary_file = self.package_source / self.binary_rel
        self.binary_file.parent.mkdir(parents=True, exist_ok=True)
        self.binary_file.write_text("binary")

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    def test_move_to_packages_success(self) -> None:
        result = files.move_to_packages(
            package_source=self.package_source,
            package_base_folder=self.package_base_folder,
            bin_base_folder=self.bin_base_folder,
            binary_path=self.binary_file,
            binary_name="symlink",
        )
        package_path = self.package_base_folder / self.package_source.name
        self.assertTrue(package_path.is_dir())
        dest_bin = self.bin_base_folder / "symlink"
        self.assertTrue(dest_bin.is_symlink())
        expected_target = package_path / self.binary_file.relative_to(self.package_source)
        self.assertEqual(os.readlink(dest_bin), str(expected_target))
        self.assertEqual(result, dest_bin)

    def test_move_to_packages_existing_package(self) -> None:
        existing = self.package_base_folder / self.package_source.name
        existing.mkdir(parents=True, exist_ok=True)
        (existing / "old.txt").write_text("old")
        _result = files.move_to_packages(
            package_source=self.package_source,
            package_base_folder=self.package_base_folder,
            bin_base_folder=self.bin_base_folder,
            binary_path=self.binary_file,
            binary_name="symlink",
        )
        package_path = self.package_base_folder / self.package_source.name
        self.assertTrue(package_path.is_dir())
        self.assertFalse((package_path / "old.txt").exists())

    def test_move_to_packages_mkdir_permission_error(self) -> None:
        with (
            patch("pathlib.Path.mkdir", side_effect=PermissionError),
            patch("sys.exit", side_effect=SystemExit) as mock_exit,
            patch.object(logger, "error") as mock_logger_error,
        ):
            with self.assertRaises(SystemExit):
                files.move_to_packages(
                    package_source=self.package_source,
                    package_base_folder=self.package_base_folder,
                    bin_base_folder=self.bin_base_folder,
                    binary_path=self.binary_file,
                    binary_name="symlink",
                )
            mock_logger_error.assert_called_once_with(
                SubstringMatcher(containing="Permission denied when creating packages folder")
            )
            mock_exit.assert_called_once_with(1)

    def test_move_to_packages_rename_permission_error(self) -> None:
        with (
            patch("pathlib.Path.rename", side_effect=PermissionError),
            patch("sys.exit", side_effect=SystemExit) as mock_exit,
            patch.object(logger, "error") as mock_logger_error,
        ):
            with self.assertRaises(SystemExit):
                files.move_to_packages(
                    package_source=self.package_source,
                    package_base_folder=self.package_base_folder,
                    bin_base_folder=self.bin_base_folder,
                    binary_path=self.binary_file,
                    binary_name="symlink",
                )
            mock_logger_error.assert_called_once_with(
                SubstringMatcher(containing="Permission denied to move package to")
            )
            mock_exit.assert_called_once_with(1)

    def test_move_to_packages_rename_oserror(self) -> None:
        self.package_base_folder.mkdir(exist_ok=True)
        with (
            patch("pathlib.Path.rename", side_effect=OSError),
            patch("sys.exit", side_effect=SystemExit) as mock_exit,
            patch.object(logger, "error") as mock_logger_error,
        ):
            with self.assertRaises(SystemExit):
                files.move_to_packages(
                    package_source=self.package_source,
                    package_base_folder=self.package_base_folder,
                    bin_base_folder=self.bin_base_folder,
                    binary_path=self.binary_file,
                    binary_name="symlink",
                )
            mock_logger_error.assert_called_once_with(
                f"Move package to '{self.package_base_folder / self.package_source.name}' failed with error: None"
            )
            mock_exit.assert_called_once_with(1)

    def test_move_to_packages_unlink_permission_error(self) -> None:
        with (
            patch("pathlib.Path.unlink", side_effect=PermissionError),
            patch("sys.exit", side_effect=SystemExit) as mock_exit,
            patch.object(logger, "error") as mock_logger_error,
        ):
            with self.assertRaises(SystemExit):
                files.move_to_packages(
                    package_source=self.package_source,
                    package_base_folder=self.package_base_folder,
                    bin_base_folder=self.bin_base_folder,
                    binary_path=self.binary_file,
                    binary_name="symlink",
                )
            mock_logger_error.assert_called_once_with(
                SubstringMatcher(containing="Permission denied to create symlink")
            )
            mock_exit.assert_called_once_with(1)

    def test_move_to_packages_symlink_permission_error(self) -> None:
        with (
            patch("pathlib.Path.symlink_to", side_effect=PermissionError),
            patch("sys.exit", side_effect=SystemExit) as mock_exit,
            patch.object(logger, "error") as mock_logger_error,
        ):
            with self.assertRaises(SystemExit):
                files.move_to_packages(
                    package_source=self.package_source,
                    package_base_folder=self.package_base_folder,
                    bin_base_folder=self.bin_base_folder,
                    binary_path=self.binary_file,
                    binary_name="symlink",
                )
            mock_logger_error.assert_called_once_with(
                SubstringMatcher(containing="Permission denied to create symlink")
            )
            mock_exit.assert_called_once_with(1)

    def test_move_to_packages_symlink_os_error(self) -> None:
        with (
            patch("pathlib.Path.symlink_to", side_effect=OSError),
            patch("sys.exit", side_effect=SystemExit) as mock_exit,
            patch.object(logger, "error") as mock_logger_error,
        ):
            with self.assertRaises(SystemExit):
                files.move_to_packages(
                    package_source=self.package_source,
                    package_base_folder=self.package_base_folder,
                    bin_base_folder=self.bin_base_folder,
                    binary_path=self.binary_file,
                    binary_name="symlink",
                )
            mock_logger_error.assert_called_once_with(SubstringMatcher(containing="Symlink moved package to"))
            mock_exit.assert_called_once_with(1)


if __name__ == "__main__":
    unittest.main()



import tempfile
import unittest
from pathlib import Path
from unittest.mock import patch

# Import the module using the new submodule path
from girsh.core import files

# -------------------------------
# Tests for get_common_prefix
# -------------------------------


class CommonPrefixTest(unittest.TestCase):
    def test_get_common_prefix_common(self) -> None:
        names = ["folder/file1.txt", "folder/file2.txt"]
        prefix = files.get_common_prefix(names)
        self.assertEqual(prefix, "folder")

    def test_get_common_prefix_no_common(self) -> None:
        names = ["folder1/file.txt", "folder2/file.txt"]
        prefix = files.get_common_prefix(names)
        self.assertIsNone(prefix)

    def test_get_common_prefix_empty(self) -> None:
        prefix = files.get_common_prefix([])
        self.assertIsNone(prefix)


# -------------------------------
# Tests for is_safe_path
# -------------------------------


class IsSafePathTest(unittest.TestCase):
    def setUp(self) -> None:
        self.temp_dir = tempfile.TemporaryDirectory()
        self.temp_path = Path(self.temp_dir.name)

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    def test_is_safe_path_true(self) -> None:
        subdir = self.temp_path / "subdir"
        subdir.mkdir()
        target = subdir / "file.txt"
        target.write_text("content")
        self.assertTrue(files.is_safe_path(self.temp_path, target))

    def test_is_safe_path_false(self) -> None:
        try:
            other_dir = tempfile.TemporaryDirectory()
            other_path = Path(other_dir.name)
            target = other_path / "file.txt"
            target.write_text("content")
            self.assertFalse(files.is_safe_path(self.temp_path, target))
        finally:
            other_dir.cleanup()

    def test_is_safe_path_oserror(self) -> None:
        target = self.temp_path / "file.txt"
        target.write_text("content")
        with patch("pathlib.Path.resolve", side_effect=OSError):
            self.assertFalse(files.is_safe_path(self.temp_path, target))


if __name__ == "__main__":
    unittest.main()



import tempfile
import unittest
from pathlib import Path
from typing import Any
from unittest.mock import patch

from loguru import logger

from girsh.core import repos


class SubstringMatcher:
    def __init__(self, containing: str) -> None:
        self.containing: str = containing.lower()

    def __eq__(self, other: Any) -> Any:
        return other.lower().find(self.containing) > -1

    def __unicode__(self) -> str:
        return f'a string containing "{self.containing}"'

    def __str__(self) -> str:
        return self.__unicode__().encode("utf-8").decode("utf-8")

    __repr__ = __unicode__


# -------------------------------
# Tests for remove_binary
# -------------------------------
class TestRemoveBinary(unittest.TestCase):
    def setUp(self) -> None:
        self.temp_dir = tempfile.TemporaryDirectory()
        self.temp_path: Path = Path(self.temp_dir.name)
        self.bin_dir = self.temp_path / "bin"
        self.bin_dir.mkdir()
        self.installed = {
            "owner/repo1": {"tag": "v1.0", "binary": "binary1", "path": str(self.bin_dir)},
            "owner/repo2": {"tag": "v1.0", "binary": "binary2", "path": str(self.bin_dir)},
        }

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    def test_remove_binary_success(self) -> None:
        binary1 = self.bin_dir / "binary1"
        binary1.touch()
        with (
            patch.object(logger, "info") as mock_logger,
        ):
            repos.remove_binary("owner/repo1", self.installed["owner/repo1"], dry_run=False)
            self.assertFalse(binary1.exists())
            mock_logger.assert_called_once_with(SubstringMatcher(containing="Uninstalled binary1"))

    def test_remove_binary_permission_error(self) -> None:
        binary1 = self.bin_dir / "binary1"
        binary1.touch()
        with (
            patch("pathlib.Path.unlink", side_effect=PermissionError),
            patch.object(logger, "error") as mock_logger,
        ):
            repos.remove_binary("owner/repo1", self.installed["owner/repo1"], dry_run=False)
            self.assertTrue(binary1.exists())
            mock_logger.assert_called_once_with(SubstringMatcher(containing="Permission denied"))

    def test_remove_binary_dry_run(self) -> None:
        binary1 = self.bin_dir / "binary1"
        binary1.touch()
        with (
            patch.object(logger, "info") as mock_logger,
        ):
            repos.remove_binary("owner/repo1", self.installed["owner/repo1"], dry_run=True)
            self.assertTrue(binary1.exists())
            mock_logger.assert_called_once_with(SubstringMatcher(containing="Dry-run:"))

    def test_remove_binary_not_exists(self) -> None:
        with (
            patch.object(logger, "info") as mock_logger,
        ):
            repos.remove_binary("owner/repo2", self.installed["owner/repo2"], dry_run=False)
        mock_logger.assert_not_called()


# -------------------------------
# Tests for remove_package
# -------------------------------
class TestRemovePackage(unittest.TestCase):
    def setUp(self) -> None:
        self.temp_dir = tempfile.TemporaryDirectory()
        self.temp_path: Path = Path(self.temp_dir.name)
        self.package_path = self.temp_path / "bin"
        self.package_path.mkdir()
        self.installed = {
            "owner/repo1": {"tag": "v1.0", "package_path": str(self.package_path / "package1")},
            "owner/repo2": {"tag": "v1.0", "package_path": str(self.package_path / "package2")},
        }

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    def test_remove_package_success(self) -> None:
        package1 = self.package_path / "package1"
        package1.mkdir(exist_ok=True)
        with (
            patch.object(logger, "info") as mock_logger,
        ):
            repos.remove_package("owner/repo1", self.installed["owner/repo1"], dry_run=False)
            self.assertFalse(package1.exists())
            mock_logger.assert_called_once_with(SubstringMatcher(containing="Deleted package folder"))

    def test_remove_package_permission_error(self) -> None:
        package1 = self.package_path / "package1"
        package1.mkdir(exist_ok=True)
        with (
            patch("shutil.rmtree", side_effect=PermissionError),
            patch.object(logger, "error") as mock_logger,
        ):
            repos.remove_package("owner/repo1", self.installed["owner/repo1"], dry_run=False)
            self.assertTrue(package1.exists())
            mock_logger.assert_called_once_with(SubstringMatcher(containing="Permission denied"))

    def test_remove_package_dry_run(self) -> None:
        package1 = self.package_path / "package1"
        package1.mkdir(exist_ok=True)
        with (
            patch.object(logger, "info") as mock_logger,
        ):
            repos.remove_package("owner/repo1", self.installed["owner/repo1"], dry_run=True)
            self.assertTrue(package1.exists())
            mock_logger.assert_called_once_with(SubstringMatcher(containing="Dry-run:"))

    def test_remove_package_not_exists(self) -> None:
        with (
            patch.object(logger, "info") as mock_logger,
        ):
            repos.remove_package("owner/repo2", self.installed["owner/repo2"], dry_run=False)
        mock_logger.assert_not_called()

    def test_remove_package_path_none(self) -> None:
        with (
            patch.object(logger, "info") as mock_logger,
        ):
            repos.remove_package("owner/repo2", {}, dry_run=False)
        mock_logger.assert_not_called()


if __name__ == "__main__":
    unittest.main()



import tempfile
import unittest
from pathlib import Path
from unittest.mock import patch

from loguru import logger

# Import the module using the new submodule path
from girsh.core import files

# -------------------------------
# Tests for find_binary
# -------------------------------


class FindBinaryTest(unittest.TestCase):
    def setUp(self) -> None:
        logger.remove()  # Remove all loguru handlers to silence logging output
        self.temp_dir: tempfile.TemporaryDirectory = tempfile.TemporaryDirectory()
        self.temp_path: Path = Path(self.temp_dir.name)
        self.bin_dir = self.temp_path / "bin"
        self.bin_dir.mkdir()
        self.app1 = self.bin_dir / "app1"
        self.app1.write_text("binary1")
        self.app1.chmod(0o755)
        self.app2 = self.bin_dir / "app2"
        self.app2.write_text("binary2")
        self.app2.chmod(0o755)

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    def test_find_binary_with_pattern(self) -> None:
        result = files.find_binary(self.bin_dir, "app1")
        self.assertEqual(result, self.app1)

    def test_find_binary_no_pattern_multiple_executables(self) -> None:
        result = files.find_binary(self.bin_dir, None)
        self.assertIn(result, [self.app1, self.app2])

    def test_find_binary_no_executables(self) -> None:
        extract_dir = self.temp_path / "empty_extract"
        extract_dir.mkdir()
        no_exec = extract_dir / "not_exec"
        no_exec.write_text("dummy_text")
        no_exec.chmod(0o644)
        result = files.find_binary(extract_dir, None)
        self.assertIsNone(result)

    def test_find_binary_empty_extract_dir(self) -> None:
        extract_dir = self.temp_path / "empty_extract"
        dummy_dir = extract_dir / "dummy"
        dummy_dir.mkdir(parents=True)
        with patch.object(logger, "debug") as mock_logger_debug:
            result = files.find_binary(extract_dir, "DummyPattern")
            mock_logger_debug.assert_called_with(f"Binary not found in {extract_dir}")
            self.assertIsNone(result)


# -------------------------------
# Tests for copy_to_bin
# -------------------------------


class CopyToBinTest(unittest.TestCase):
    def setUp(self) -> None:
        logger.remove()  # Remove all loguru handlers to silence logging output
        self.temp_dir: tempfile.TemporaryDirectory = tempfile.TemporaryDirectory()
        self.temp_path: Path = Path(self.temp_dir.name)
        self.binary_file = self.temp_path / "binary"
        self.binary_file.write_text("binary")
        self.bin_base_folder = self.temp_path / "bin"
        self.bin_base_folder.mkdir()

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    def test_copy_to_bin_success(self) -> None:
        result = files.copy_to_bin(
            binary_path=self.binary_file, bin_base_folder=self.bin_base_folder, binary_name="new_binary"
        )
        dest = self.bin_base_folder / "new_binary"
        self.assertEqual(result, dest)
        self.assertTrue(dest.exists())
        self.assertEqual(dest.read_text(), "binary")
        self.assertEqual(dest.stat().st_mode & 0o111, 0o111)

    def test_copy_to_bin_permission_error(self) -> None:
        # Patch the mkdir method on the entire Path class.
        with patch("pathlib.Path.mkdir", side_effect=PermissionError), self.assertRaises(SystemExit):
            files.copy_to_bin(
                binary_path=self.binary_file,
                bin_base_folder=self.bin_base_folder,
                binary_name="new_binary",
            )

    def test_copy_to_bin_copy_permission_error(self) -> None:
        with patch("shutil.copy2", side_effect=PermissionError), self.assertRaises(SystemExit):
            files.copy_to_bin(
                binary_path=self.binary_file,
                bin_base_folder=self.bin_base_folder,
                binary_name="new_binary",
            )

    def test_copy_to_bin_copy_oserror(self) -> None:
        with patch("shutil.copy2", side_effect=OSError("error")), self.assertRaises(SystemExit):
            files.copy_to_bin(
                binary_path=self.binary_file,
                bin_base_folder=self.bin_base_folder,
                binary_name="new_binary",
            )


if __name__ == "__main__":
    unittest.main()



import tempfile
import unittest
from dataclasses import dataclass
from pathlib import Path
from typing import Any
from unittest.mock import MagicMock, patch

from loguru import logger

from girsh.core import repos


class SubstringMatcher:
    def __init__(self, containing: str) -> None:
        self.containing: str = containing.lower()

    def __eq__(self, other: Any) -> Any:
        return other.lower().find(self.containing) > -1

    def __unicode__(self) -> str:
        return f'a string containing "{self.containing}"'

    def __str__(self) -> str:
        return self.__unicode__().encode("utf-8").decode("utf-8")

    __repr__ = __unicode__


# -------------------------------
# Dummy classes used for testing
# -------------------------------
@dataclass
class DummyRepository:
    version: str | None = None
    comment: str = ""
    package_pattern: str | None = None
    download_url: str | None = None
    filter_pattern: str | None = None
    multi_file: bool = False
    binary_name: str | None = None
    pre_update_commands: list[str] | None = None
    post_update_commands: list[str] | None = None


def dummy_show_installed(installed: dict[Any, Any], repositories: dict) -> int:
    return 0


def dummy_remove_binary(repo: str, info: dict, dry_run: bool) -> None:
    return


def dummy_remove_package(repo: str, info: dict, dry_run: bool) -> None:
    return


def dummy_save_installed(installed_file: Path, data: dict) -> int:
    return 42


# -------------------------------
# Tests for uninstall
# -------------------------------
class TestUninstall(unittest.TestCase):
    def setUp(self) -> None:
        # Disable loguru logger output
        from loguru import logger

        logger.remove()
        self.temp_dir = tempfile.TemporaryDirectory()
        self.temp_path: Path = Path(self.temp_dir.name)
        self.bin_dir = self.temp_path / "bin"
        self.dummy_repositories = {"dummy1": DummyRepository(version="v1.0")}
        self.dummy_installed: dict[Any, Any] = {}

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    @patch("girsh.core.repos.logger")
    def test_uninstall_not_installed(self, mock_logger: MagicMock) -> None:
        result = repos.uninstall(
            repositories=self.dummy_repositories,  # type: ignore[arg-type]
            installed={},
            dry_run=False,
        )
        self.assertEqual(result, {})
        mock_logger.log.assert_called_once_with("SUCCESS", "No binaries installed.")

    def test_uninstall_removes_unlisted_repo(self) -> None:
        binary_name = "binary1"
        pre_commands = ["echo UNINSTALL", rf"%stop_processes% {binary_name}"]
        installed = {
            "unknown/repo": {
                "tag": "v1.0",
                "binary": binary_name,
                "path": str(self.temp_path),
                "pre_update_commands": pre_commands,
            }
        }
        binary_path = self.temp_path / binary_name
        binary_path.touch()
        with (
            patch.object(logger, "success") as mock_logger_success,
            patch.object(logger, "info") as mock_logger_info,
            patch("girsh.core.utils.run_commands", return_value=True) as mock_run_commands,
        ):
            summary: dict = repos.uninstall(
                repositories=self.dummy_repositories,  # type: ignore[arg-type]
                installed=installed,
                dry_run=False,
            )
        self.assertEqual(summary, {repos.RepoResult.uninstalled: 1})
        self.assertNotIn("unknown/repo", installed)
        mock_run_commands.assert_called_once_with(pre_commands, "repo: pre-uninstall")
        info_logs = [
            "Uninstall binaries no longer in config.",
            "Running commands for repo: pre-uninstall",
        ]
        for n, call in enumerate(mock_logger_info.call_args_list):
            self.assertEqual(call.args[0], info_logs[n])

        mock_logger_success.assert_called_once_with(
            f"Uninstalled {binary_path.name} from {binary_path} originated from unknown/repo"
        )

    def test_uninstall_all(self) -> None:
        binary_name = "binary1"
        pre_commands = ["echo UNINSTALL", rf"%stop_processes% {binary_name}"]
        installed = {
            "unknown/repo": {
                "tag": "v1.0",
                "binary": binary_name,
                "path": str(self.temp_path),
                "pre_update_commands": pre_commands,
            }
        }
        binary_path = self.temp_path / binary_name
        binary_path.touch()
        with (
            patch.object(logger, "success") as mock_logger_success,
            patch.object(logger, "info") as mock_logger_info,
            patch("girsh.core.utils.run_commands", return_value=True) as mock_run_commands,
            patch("girsh.core.utils.confirm_default_no", return_value=True),
        ):
            summary: dict = repos.uninstall(
                repositories=[],  # Uninstall-all is achieved by passing empty repositories dict
                installed=installed,
                dry_run=False,
            )
        self.assertEqual(summary, {repos.RepoResult.uninstalled: 1})
        self.assertNotIn("unknown/repo", installed)
        mock_run_commands.assert_called_once_with(pre_commands, "repo: pre-uninstall")
        info_logs = [
            "Uninstall all binaries.",
            "Running commands for repo: pre-uninstall",
        ]
        for n, call in enumerate(mock_logger_info.call_args_list):
            self.assertEqual(call.args[0], info_logs[n])

        mock_logger_success.assert_called_once_with(
            f"Uninstalled {binary_path.name} from {binary_path} originated from unknown/repo"
        )

    def test_uninstall_all_cancelled(self) -> None:
        binary_name = "binary1"
        pre_commands = ["echo UNINSTALL", rf"%stop_processes% {binary_name}"]
        installed = {
            "unknown/repo": {
                "tag": "v1.0",
                "binary": binary_name,
                "path": str(self.temp_path),
                "pre_update_commands": pre_commands,
            }
        }
        binary_path = self.temp_path / binary_name
        binary_path.touch()
        with (
            patch.object(logger, "success") as mock_logger_success,
            patch.object(logger, "info") as mock_logger_info,
            patch("girsh.core.utils.run_commands", return_value=True) as mock_run_commands,
            patch("girsh.core.utils.confirm_default_no", return_value=False),
        ):
            summary: dict = repos.uninstall(
                repositories=[],  # Uninstall-all is achieved by passing empty repositories dict
                installed=installed,
                dry_run=False,
            )
        self.assertEqual(summary, {repos.RepoResult.cancelled: 1})
        self.assertIn("unknown/repo", installed)
        mock_run_commands.assert_not_called()
        info_logs = [
            "Uninstall cancelled.",
            "Running commands for repo: pre-uninstall",
        ]
        for n, call in enumerate(mock_logger_info.call_args_list):
            self.assertEqual(call.args[0], info_logs[n])
        mock_logger_success.assert_not_called()

    def test_uninstall_not_exists(self) -> None:
        installed = {"unknown/repo": {"tag": "v1.0", "binary": "binary1", "path": str(self.temp_path)}}
        # binary_path = self.temp_path / installed["unknown/repo"]["binary"]
        # # binary_path.touch()
        with patch.object(logger, "info") as mock_logger:
            summary: dict = repos.uninstall(
                repositories=self.dummy_repositories,  # type: ignore[arg-type]
                installed=installed,
                dry_run=False,
            )
        self.assertEqual(summary, {repos.RepoResult.uninstalled: 1})
        self.assertNotIn("unknown/repo", installed)
        mock_logger.assert_called_with("Uninstall binaries no longer in config.")

    def test_uninstall_keep_listed_repo(self) -> None:
        installed = {"dummy1": {"tag": "v1.0", "binary": "binary1", "path": str(self.temp_path)}}
        with patch.object(logger, "info") as mock_logger:
            summary: dict = repos.uninstall(
                repositories=self.dummy_repositories,  # type: ignore[arg-type]
                installed=installed,
                dry_run=False,
            )
        self.assertEqual(summary, {})
        self.assertNotIn("unknown/repo", installed)
        mock_logger.assert_called_with("Uninstall binaries no longer in config.")

    def test_uninstall_dry_run(self) -> None:
        installed = {"unknown/repo": {"tag": "v1.0", "binary": "binary1", "path": str(self.temp_path)}}
        binary_path = self.temp_path / installed["unknown/repo"]["binary"]
        binary_path.touch()
        with (
            patch.object(logger, "success") as mock_logger,
        ):
            result = repos.uninstall(
                repositories=self.dummy_repositories,  # type: ignore[arg-type]
                installed=installed,
                dry_run=True,
            )
        self.assertEqual(result, {repos.RepoResult.dry_run_uninstall: 1})
        mock_logger.assert_called_with(f"Dry-run: unknown/repo: Would uninstall binary1 from {binary_path}.")

    def test_uninstall_permission_error(self) -> None:
        installed = {"unknown/repo": {"tag": "v1.0", "binary": "binary1", "path": str(self.temp_path)}}
        binary_path = self.temp_path / installed["unknown/repo"]["binary"]
        binary_path.touch()
        with (
            patch("pathlib.Path.unlink", side_effect=PermissionError),
            patch("girsh.core.utils.run_commands", return_value=True),
            patch.object(logger, "error") as mock_logger,
        ):
            result = repos.uninstall(
                repositories=self.dummy_repositories,  # type: ignore[arg-type]
                installed=installed,
                dry_run=False,
            )
        self.assertEqual(result, {repos.RepoResult.uninstall_failed: 1})
        mock_logger.assert_called_with(f"Permission denied when trying to uninstall {binary_path}")


if __name__ == "__main__":
    unittest.main()



import tempfile
import unittest
from pathlib import Path
from typing import Any
from unittest.mock import Mock, patch

import requests
from loguru import logger
from requests.exceptions import HTTPError, RequestException

# Import the module using the new submodule path
from girsh.core import files

# -------------------------------
# Tests for download_github_release
# -------------------------------


class DownloadGithubReleaseTest(unittest.TestCase):
    def setUp(self) -> None:
        logger.remove()  # Remove all loguru handlers to silence logging output
        self.temp_dir: tempfile.TemporaryDirectory = tempfile.TemporaryDirectory()
        self.output_dir: Path = Path(self.temp_dir.name)
        # self.dummy_file = self.download_dir / "dummy_file.txt"
        # self.dummy_file.touch()  # Create a dummy file in the download folder

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    def test_download_github_release_with_download_url(self) -> None:
        release_info = {"tag_name": "v1.0", "assets": []}
        with patch("girsh.core.files.download_package", return_value=Path(self.output_dir) / "dummy.txt"):
            result = files.download_github_release(
                "http://api.github.com/releases/latest",
                "pattern",
                self.output_dir,
                release_info,
                download_url="http://example.com/{version}/download.zip",
            )
        self.assertIsNotNone(result)
        if result:
            path, tag = result
            self.assertEqual(tag, "v1.0")
            self.assertEqual(path.name, "dummy.txt")

    def test_download_github_release_with_asset_match(self) -> None:
        release_info = {
            "tag_name": "v2.0",
            "assets": [{"name": "match.zip", "browser_download_url": "http://example.com/match.zip"}],
        }
        fake_func = lambda url, out_dir, filename=None: Path(out_dir) / (filename if filename else "dummy")
        with patch("girsh.core.files.download_package", side_effect=fake_func):
            result = files.download_github_release(
                "http://api.github.com/releases/latest", "match", self.output_dir, release_info
            )
        self.assertIsNotNone(result)
        if result:
            path, tag = result
            self.assertEqual(tag, "v2.0")
            self.assertEqual(path.name, "match.zip")

    def test_download_github_release_with_asset_match_none(self) -> None:
        release_info = {
            "tag_name": "v2.0",
            "assets": [{"name": "match.zip"}],
        }
        fake_func = lambda url, out_dir, filename=None: Path(out_dir) / (filename if filename else "dummy")
        with (
            patch("girsh.core.files.download_package", side_effect=fake_func),
            patch.object(logger, "warning") as mock_logger,
        ):
            result = files.download_github_release(
                "http://api.github.com/releases/latest", "match", self.output_dir, release_info
            )
            mock_logger.assert_called_once_with("No download URL found for asset: match.zip")
        self.assertIsNone(result)

    def test_download_github_release_no_match(self) -> None:
        release_info = {
            "tag_name": "v3.0",
            "assets": [{"name": "nomatch.zip", "browser_download_url": "http://example.com/nomatch.zip"}],
        }
        result = files.download_github_release(
            "http://api.github.com/releases/latest", "pattern", self.output_dir, release_info
        )
        self.assertIsNone(result)

    def test_download_github_release_http_error(self) -> None:
        def fake_get(url: str, headers: dict[str, str], timeout: int) -> Any:
            raise HTTPError("error")

        with patch("requests.get", side_effect=fake_get):
            result = files.download_github_release("http://api.github.com/releases/latest", "pattern", self.output_dir)
        self.assertIsNone(result)

    def test_download_github_release_request_exception(self) -> None:
        def fake_get(url: str, headers: dict[str, str], timeout: int) -> Any:
            raise RequestException("error")

        with patch("requests.get", side_effect=fake_get):
            result = files.download_github_release("http://api.github.com/releases/latest", "pattern", self.output_dir)
        self.assertIsNone(result)

    @patch("girsh.core.files.requests.get")
    def girsh(self, mock_get: Mock) -> None:
        """Test case when raise_for_status() raises an HTTPError."""
        # Mock the response object
        mock_response = Mock()
        mock_get.return_value = mock_response

        # Set up the mock for raise_for_status to raise an HTTPError
        mock_response.raise_for_status.side_effect = requests.HTTPError("HTTP Error")

        # Provide dummy parameters for the function
        url: str = "https://api.github.com/repos/user/repo/releases/latest"
        package_pattern: str = "some_pattern"

        # Call the function and verify if raise_for_status was called and exception is raised
        with self.assertRaises(requests.HTTPError):
            files.download_github_release(url, package_pattern, self.output_dir)

    @patch("girsh.core.files.requests.get")
    def test_download_github_release_success(self, mock_get: Mock) -> None:
        """Test case for a successful download."""
        # Mock the response object for a successful request
        mock_response = Mock()
        mock_get.return_value = mock_response

        # Simulate the response being successful and return JSON data
        mock_response.raise_for_status.return_value = None  # No exception raised
        mock_response.json.return_value = {
            "tag_name": "v1.0.0",
            "assets": [{"name": "asset1.zip", "browser_download_url": "http://example.com/asset1.zip"}],
        }

        # Mock the `iter_content` method of the response to simulate downloading in chunks
        mock_response.iter_content.return_value = [b"chunk1", b"chunk2", b"chunk3"]

        # Provide dummy parameters for the function
        url: str = "https://api.github.com/repos/user/repo/releases/latest"
        package_pattern: str = "asset1"

        # Call the function and assert the expected return value
        result = files.download_github_release(url, package_pattern, self.output_dir)
        self.assertIsNotNone(result)
        self.assertEqual(result[1], "v1.0.0")  # type: ignore[index]


if __name__ == "__main__":
    unittest.main()



import argparse
import os
import subprocess
import sys
from collections import defaultdict
from dataclasses import dataclass
from importlib import resources
from importlib.metadata import version
from pathlib import Path
from types import UnionType
from typing import Any, get_args, get_origin

import yaml
from loguru import logger

__version__ = version("girsh")


class ConversionError(TypeError):
    def __init__(self, value: Any, expected_type: type) -> None:
        super().__init__(f"Cannot convert {value!r} to {expected_type.__name__}")
        self.value = value
        self.expected_type = expected_type


def convert_to_bool(value: Any) -> Any:
    if isinstance(value, bool):
        return value
    elif isinstance(value, str) and value.lower() in ["true", "false"]:
        return value.lower() == "true"
    else:
        raise ValueError


@dataclass
class General:
    bin_base_folder: Path = Path("/usr/local/bin") if os.geteuid() == 0 else Path.home() / ".local/bin"
    installed_file: Path = (
        Path("/etc/girsh/installed.yaml") if os.geteuid() == 0 else Path.home() / ".config" / "girsh" / "installed.yaml"
    )
    download_dir: Path = Path.home() / ".cache" / "girsh" / "downloads"
    package_pattern: str = r".*x86_64.*(gz|zip)$"  # Regex to find the desired package in the release assets
    package_base_folder: Path = Path("/opt/girsh") if os.geteuid() == 0 else Path.home() / ".local/share/girsh"

    def __setattr__(self, name: str, value: Any) -> None:
        expected_type = self.__annotations__.get(name)
        if expected_type is not None and not isinstance(value, expected_type):
            try:
                value = expected_type(value)
            except (TypeError, ValueError) as e:
                raise ConversionError(value, expected_type) from e
        super().__setattr__(name, value)


@dataclass
class Repository:
    comment: str = ""  # Description of the package
    package_pattern: str | None = None  # Regex to find the desired package in the release assets
    filter_pattern: str | None = None  # Regex to find the program file in extracted files
    binary_name: str | None = None  # Rename extracted binary file
    multi_file: bool = False  # Package requires all package files (not single binary)
    version: str | None = None  # Pin program to defined version
    download_url: str | None = None  # Optional download URL template, can contain `{version}` as a variable
    pre_update_commands: list[str | None] | None = None  # Command to run before updating the package
    post_update_commands: list[str | None] | None = None  # Command to run after updating the package

    def __setattr__(self, name: str, value: Any) -> None:
        expected_type: type | None = self.__annotations__.get(name)
        expected_type = get_args(expected_type)[0] if get_origin(expected_type) is UnionType else expected_type
        if get_origin(expected_type) is not None:  # type is a parameterized generic
            expected_type = get_origin(expected_type)
        logger.trace(f"Setting {name} to {value} of type {expected_type}")
        if value is not None and expected_type is not None and not isinstance(value, expected_type):
            try:
                value = convert_to_bool(value) if expected_type is bool else expected_type(value)
            except (TypeError, ValueError) as e:
                raise ConversionError(value, expected_type) from e
        super().__setattr__(name, value)


def update_general_config(general_config: General, data: dict) -> General:
    """
    Update a General dataclass instance with values from a dictionary.

    This function looks for a 'general' key in the provided data and updates
    the configuration fields if they exist, triggering any type conversions
    defined in __setattr__.

    Args:
        general_config (General): The current configuration instance.
        data (dict): Dictionary containing configuration data, e.g., from a YAML file.

    Returns:
        General: The updated configuration instance.
    """
    general_updates = data.get("general")
    if not isinstance(general_updates, dict):
        if general_updates is not None:
            logger.error(f"General config format is not dict: {data}")
        return general_config

    for key, value in general_updates.items():
        if key in general_config.__annotations__:
            setattr(general_config, key, value)
        else:
            logger.warning(f"Skipped general config entry {key}: {value}")
    return general_config


def update_repositories_config(
    repo_config: dict[str, Repository], data: dict, default_pattern: str
) -> dict[str, Repository]:
    """
    Update repository configurations using YAML data.

    This function extracts repository configurations from the "repositories" key in the provided
    YAML data. For each repository, it creates a new Repository instance (or a default one if the
    configuration is None) and ensures that a default package pattern is set if missing.

    Args:
        repo_config (dict[str, Repository]): Existing repository configuration mapping.
        data (dict): Loaded configuration data from a YAML file.
        default_pattern (str): Default package pattern to use when not specified.

    Returns:
        dict[str, Repository]: The updated repository configuration mapping.
    """
    repositories_data = data.get("repositories")
    if repositories_data is None:
        logger.warning(f"No repositories configured.{data}")
        return repo_config

    if not isinstance(repositories_data, dict):
        logger.error("The 'repositories' YAML config is not a dictionary.")
        return repo_config

    for repo_name, repo_data in repositories_data.items():
        try:
            repository = Repository(**repo_data) if repo_data is not None else Repository()
        except TypeError as e:
            logger.error(f"Repository '{repo_name}': {e}")
            continue
        if repository.package_pattern is None:
            repository.package_pattern = default_pattern

        repo_config[repo_name] = repository

    return repo_config


def load_yaml_config(file_path: str) -> tuple[General, dict[str, Repository]]:
    """
    Load and parse the YAML configuration file and update the general and repositories config.

    Args:
        file_path (str): Path to the YAML configuration file.

    Returns:
        tuple[General, dict[str, Repository]]: A tuple containing the updated General instance
            and a dictionary mapping repository names to Repository instances.
    """
    config_path = Path(file_path)
    try:
        with config_path.open("r") as file:
            data = yaml.safe_load(file)
            if not isinstance(data, dict):
                logger.warning(f"No valid configuration data found in '{file_path}'.")
                return General(), {}
    except FileNotFoundError:
        logger.error(f"Config file '{file_path}' not found.")
        sys.exit(1)
    except PermissionError:
        logger.error(f"Permission denied when reading config file '{file_path}'.")
        sys.exit(1)
    except (yaml.scanner.ScannerError, yaml.parser.ParserError) as err:
        logger.error(f"YAML syntax error in config file '{file_path}': {err}")
        sys.exit(1)

    logger.trace(f"Loaded config: {data}")
    general = update_general_config(General(), data)
    logger.trace(f"General config: {general}")

    repositories: dict[str, Repository] = defaultdict(Repository)
    repositories = update_repositories_config(repositories, data, general.package_pattern)
    logger.trace(f"Repositories config: {repositories}")

    # Check if there are other unexpected keys in the yaml config
    unexpected = set(data.keys()) - {"general", "repositories"}
    if unexpected:
        logger.warning(f"Config YAML contains unexpected keys: {unexpected}")

    return general, repositories


def edit_config(config_path: Path) -> int:
    """
    Open the specified config file in the user's default terminal editor.

    Args:
        config_path (Path): Path to the configuration file to edit.

    Returns:
        int: error code
    """
    # Ensure the config file exists or ask to create an empty one if not found
    if not config_path.exists():
        create_file = (
            input(f"The file '{config_path}' does not exist. Do you want to create it? (y/N): ").strip().lower()
        )
        if create_file == "y":
            logger.info("Create config file from template")
            config_path.parent.mkdir(parents=True, exist_ok=True)
            with resources.path("girsh.templates", "config_template.yaml") as template_path:
                config_content = template_path.read_text()
                config_path.write_text(config_content)
        else:
            logger.error("Operation aborted. No file was created.")
            return 0

    # Record the file's last modification time
    original_mod_time = config_path.stat().st_mtime

    # Get the user's default editor from environment variables
    editor = os.environ.get("EDITOR") or os.environ.get("VISUAL") or "nano"

    try:
        # Open the file in the chosen editor
        subprocess.run([editor, str(config_path)], check=True)  # noqa: S603
        # Check if the file modification time changed after editing
        updated_mod_time = config_path.stat().st_mtime
        if updated_mod_time != original_mod_time:
            logger.info("The config file was modified.")
        else:
            logger.info("No changes were made to the config file.")
    except FileNotFoundError:
        logger.error(f"Error: Editor '{editor}' not found. Please set the EDITOR environment variable.")
        return 1
    except subprocess.CalledProcessError as e:
        logger.error(f"Error: Failed to open {config_path} in {editor}. {e!s}")
        return 1
    return 0


def get_arguments() -> argparse.Namespace:
    """
    Parse the command line arguments

    Returns:
        argparse.Namespace: Parsed command line arguments
    """
    parser = argparse.ArgumentParser(description="Git Install Released Software Helper")

    # Add mutually exclusive group for commands, default command is install/update
    commands = parser.add_mutually_exclusive_group(required=False)
    commands.add_argument(
        "-r", "--reinstall", nargs="+", metavar="BINARY", help="Force re-installation even if version unchanged"
    )
    commands.add_argument(
        "-u",
        "--uninstall",
        action="store_true",
        help="Uninstall previously installed binary if not present in config anymore",
    )
    commands.add_argument("--uninstall-all", action="store_true", help="Uninstall all previously installed binaries")
    commands.add_argument("--clean", action="store_true", help="Remove the downloads folder and exit")
    commands.add_argument("-s", "--show", action="store_true", help="Show config and currently installed binaries")
    commands.add_argument("-e", "--edit", action="store_true", help="Open the config file in the default editor")

    parser.add_argument(
        "-c",
        "--config",
        type=Path,
        default=Path("/etc/girsh/settings.yaml")
        if os.geteuid() == 0
        else Path.home() / ".config" / "girsh" / "settings.yaml",
        help="Path to config file, defaults to ~/.config/girsh.yaml",
    )
    parser.add_argument(
        "-d", "--dry-run", action="store_true", help="Run without actually installing or removing any files."
    )
    parser.add_argument("-v", "--verbose", action="count", default=0, help="Increase output verbosity (up to 3 times)")
    parser.add_argument("-g", "--global", dest="system", action="store_true", help="Install as root at system level")
    parser.add_argument("-V", "--version", action="version", version=f"girsh {__version__}")
    return parser.parse_args()



import unittest

from loguru import logger

# Import the module using the new submodule path
from girsh.core import files

# -------------------------------
# Tests for get_filename_from_cd
# -------------------------------


class GetFilenameFromCDTest(unittest.TestCase):
    def setUp(self) -> None:
        logger.remove()  # Remove all loguru handlers to silence logging output

    def test_get_filename_from_cd_none(self) -> None:
        self.assertIsNone(files.get_filename_from_cd(None))

    def test_get_filename_from_cd_empty(self) -> None:
        self.assertIsNone(files.get_filename_from_cd(""))

    def test_get_filename_from_cd_valid(self) -> None:
        header = 'attachment; filename="archive.tar.gz"'
        filename = files.get_filename_from_cd(header)
        self.assertEqual(filename, '"archive.tar.gz"')

    def test_get_filename_from_cd_invalid(self) -> None:
        header = "attachment; something_else"
        filename = files.get_filename_from_cd(header)
        self.assertIsNone(filename)


if __name__ == "__main__":
    unittest.main()
